import os
import warnings
from collections import OrderedDict

from larray.util.misc import _positive_integer, _validate_dir
from larray.core.group import _translate_sheet_name
from larray.core.array import asarray, zip_array_items
from larray.example import load_example_data, EXAMPLE_EXCEL_TEMPLATES_DIR

try:
    import xlwings as xw
except ImportError:
    xw = None


_default_items_size = {}


def _validate_template_filename(filename):
    filename, ext = os.path.splitext(filename)
    if not ext:
        ext = '.crtx'
    if ext != '.crtx':
        raise ValueError(f"Extension for the excel template file must be '.crtx' instead of {ext}")
    return filename + ext


class AbstractReportItem:
    def __init__(self, template_dir=None, template=None, graphs_per_row=1):
        self.template_dir = template_dir
        self.template = template
        self.default_items_size = _default_items_size.copy()
        self.graphs_per_row = graphs_per_row

    @property
    def template_dir(self):
        r"""
        Set the path to the directory containing the Excel template files (with '.crtx' extension).
        This method is mainly useful if your template files are located in several directories,
        otherwise pass the template directory directly the ExcelReport constructor.

        Parameters
        ----------
        template_dir : str
            Path to the directory containing the Excel template files.

        See Also
        --------
        set_graph_template

        Examples
        --------
        >>> report = ExcelReport(EXAMPLE_EXCEL_TEMPLATES_DIR)
        >>> # ... add some graphs using template files from 'C:\excel_templates_dir'
        >>> report.template_dir = r'C:\other_templates_dir' # doctest: +SKIP
        >>> # ... add some graphs using template files from 'C:\other_templates_dir'
        """
        return self._template_dir

    @template_dir.setter
    def template_dir(self, template_dir):
        if template_dir is not None:
            _validate_dir(template_dir)
            template_dir = template_dir
        self._template_dir = template_dir

    @property
    def template(self):
        r"""
        Set a default Excel template file.

        Parameters
        ----------
        template : str
            Name of the template to be used as default template.
            The extension '.crtx' will be added if not given.
            The full path to the template file must be given if no template directory has been set.

        Examples
        --------
        >>> demo = load_example_data('demography_eurostat')

        Passing the name of the template (only if a template directory has been set)

        >>> report = ExcelReport(EXAMPLE_EXCEL_TEMPLATES_DIR)
        >>> report.template = 'Line'

        >>> sheet_population = report.new_sheet('Population')
        >>> sheet_population.add_graph(demo.population['Belgium'],'Belgium')

        Passing the full path of the template file

        >>> # if no default template directory has been set
        >>> # or if the new template is located in another directory,
        >>> # you must provide the full path
        >>> sheet_population.template = r'C:\other_templates_dir\Line_Marker.crtx' # doctest: +SKIP
        >>> sheet_population.add_graph(demo.population['Germany'],'Germany') # doctest: +SKIP
        """
        return self._template

    @template.setter
    def template(self, template):
        if template is not None:
            if self.template_dir is None:
                raise RuntimeError("Please set template_dir first")
            filename = _validate_template_filename(template)
            template = os.path.join(self.template_dir, filename)
        self._template = template

    def set_item_default_size(self, kind, width=None, height=None):
        r"""
        Override the default 'width' and 'height' values for the given kind of item.
        A new value must be provided at least for 'width' or 'height'.

        Parameters
        ----------
        kind : str
            kind of item for which default values of 'width' and/or 'height' are modified.
            Currently available kinds are 'title' and 'graph'.
        width : int, optional
            new default width value.
        height : int, optional
            new default height value.

        Examples
        --------
        >>> report = ExcelReport()
        >>> report.set_item_default_size('graph', width=450, height=250)
        """
        if width is None and height is None:
            raise ValueError("No value provided for both 'width' and 'heigth'. "
                             "Please provide one for at least 'width' or 'heigth'")
        if kind not in self.default_items_size:
            item_types = sorted(self.default_items_size.keys())
            raise ValueError(f"Item type {kind} is not registered. Please choose in list {item_types}")
        if width is None:
            width = self.default_items_size[kind].width
        if height is None:
            height = self.default_items_size[kind].height
        self.default_items_size[kind] = ItemSize(width, height)

    @property
    def graphs_per_row(self):
        r"""
        Default number of graphs per row.

        Parameters
        ----------
        graphs_per_row: int

        See Also
        --------
        ReportSheet.newline
        """
        return self._graphs_per_row

    @graphs_per_row.setter
    def graphs_per_row(self, graphs_per_row):
        _positive_integer(graphs_per_row)
        self._graphs_per_row = graphs_per_row


class AbstractReportSheet(AbstractReportItem):
    r"""
    Represents a sheet dedicated to contains only graphical items (title banners, graphs).
    See :py:obj:`ExcelReport` for use cases.

    Parameters
    ----------
    template_dir : str, optional
        Path to the directory containing the Excel template files (with a '.crtx' extension).
        Defaults to None.
    template : str, optional
        Name of the template to be used as default template.
        The extension '.crtx' will be added if not given.
        The full path to the template file must be given if no template directory has been set.
        Defaults to None.
    graphs_per_row : int, optional
        Default number of graphs per row. Defaults to 1.

    See Also
    --------
    ExcelReport
    """
    def add_title(self, title, width=None, height=None, fontsize=11):
        r"""
        Add a title item to the current sheet.
        Note that the current method only add a new item to the list of items to be generated.
        The report Excel file is generated only when the :py:obj:`~ExcelReport.to_excel` is called.

        Parameters
        ----------
        title : str
            Text to write in the title item.
        width : int, optional
            width of the title item. The current default value is used if None
            (see :py:obj:`~ExcelReport.set_item_default_size`). Defaults to None.
        height : int, optional
            height of the title item. The current default value is used if None
            (see :py:obj:`~ExcelReport.set_item_default_size`). Defaults to None.
        fontsize : int, optional
            fontsize of the displayed text. Defaults to 11.

        Examples
        --------
        >>> report = ExcelReport()

        >>> first_sheet = report.new_sheet('First_sheet')
        >>> first_sheet.add_title('Title banner with default width, height and fontsize')
        >>> first_sheet.add_title('Larger title banner', width=1200, height=100)
        >>> first_sheet.add_title('Bigger fontsize', fontsize=13)

        >>> # do not forget to call 'to_excel' to create the report file
        >>> report.to_excel('Report.xlsx')
        """
        pass

    def add_graph(self, data, title=None, template=None, width=None, height=None, min_y=None, max_y=None,
                  xticks_spacing=None, customize_func=None, customize_kwargs=None):
        r"""
        Add a graph item to the current sheet.
        Note that the current method only add a new item to the list of items to be generated.
        The report Excel file is generated only when the :py:obj:`~ExcelReport.to_excel` is called.

        Parameters
        ----------
        data : 1D or 2D array-like
            1D or 2D array representing the data associated with the graph.
            The first row represents the abscissa labels.
            Each additional row represents a new series and must start with the name of the current series.
        title : str, optional
            title of the graph. Defaults to None.
        template : str, optional
            name of the template to be used to generate the graph.
            The full path to the template file must be provided if no template directory has not been set
            or if the template file belongs to another directory.
            Defaults to the defined template (see :py:obj:`~ExcelReport.set_graph_template`).
        width : int, optional
            width of the title item. The current default value is used if None
            (see :py:obj:`~ExcelReport.set_item_default_size`). Defaults to None.
        height : int, optional
            height of the title item. The current default value is used if None
            (see :py:obj:`~ExcelReport.set_item_default_size`). Defaults to None.
        min_y: int, optional
            minimum value for the Y axis.
        max_y: int, optional
            maximum value for the Y axis.
        xticks_spacing: int, optional
            space interval between two ticks along the X axis.
        customize_func: function, optional
            user defined function to personalize the graph.
            The function must take the Chart object as first argument.
            All keyword arguments defined in customize_kwargs are passed to the function at call.
        customize_kwargs: dict, optional
            keywords arguments passed to the function `customize_func` at call.

        Examples
        --------
        >>> demo = load_example_data('demography_eurostat')
        >>> report = ExcelReport(EXAMPLE_EXCEL_TEMPLATES_DIR)

        >>> sheet_be = report.new_sheet('Belgium')

        Specifying the 'template'

        >>> sheet_be.add_graph(demo.population['Belgium'], 'Population', template='Line')

        Specifying the 'template', 'width' and 'height' values

        >>> sheet_be.add_graph(demo.births['Belgium'], 'Births', template='Line', width=450, height=250)

        Setting a default template

        >>> sheet_be.template = 'Line_Marker'
        >>> sheet_be.add_graph(demo.deaths['Belgium'], 'Deaths')

        Specify the mininum and maximum values for the Y axis

        >>> sheet_be.add_graph(demo.population['Belgium'], 'Population (min/max Y axis = 5/6 millions)', min_y=5e6, max_y=6e6)

        Specify the interval between two ticks (X axis)

        >>> sheet_be.add_graph(demo.population['Belgium'], 'Population (every 2 years)', xticks_spacing=2)

        Dumping the report Excel file

        >>> # do not forget to call 'to_excel' to create the report file
        >>> report.to_excel('Demography_Report.xlsx')
        """
        pass

    def add_graphs(self, array_per_title, axis_per_loop_variable, template=None, width=None, height=None,
                   graphs_per_row=1, min_y=None, max_y=None, xticks_spacing=None, customize_func=None,
                   customize_kwargs=None):
        r"""
        Add multiple graph items to the current sheet. This method is mainly useful when multiple
        graphs are generated by iterating over one or several axes of an array (see examples below).
        The report Excel file is generated only when the :py:obj:`~ExcelReport.to_excel` is called.

        Parameters
        ----------
        array_per_title: dict
            dictionary containing pairs (title template, array).
        axis_per_loop_variable: dict
            dictionary containing pairs (variable used in the title template, axis).
        template : str, optional
            name of the template to be used to generate the graph.
            The full path to the template file must be provided if no template directory has not been set
            or if the template file belongs to another directory.
            Defaults to the defined template (see :py:obj:`~ExcelReport.set_graph_template`).
        width : int, optional
            width of the title item. The current default value is used if None
            (see :py:obj:`~ExcelReport.set_item_default_size`). Defaults to None.
        height : int, optional
            height of the title item. The current default value is used if None
            (see :py:obj:`~ExcelReport.set_item_default_size`). Defaults to None.
        graphs_per_row: int, optional
            Number of graphs per row. Defaults to 1.
        min_y: int, optional
            minimum value for the Y axis.
        max_y: int, optional
            maximum value for the Y axis.
        xticks_spacing: int, optional
            space interval between two ticks along the X axis.
        customize_func: function, optional
            user defined function to personalize the graph.
            The function must take the Chart object as first argument.
            All keyword arguments defined in customize_kwargs are passed to the function at call.
        customize_kwargs: dict, optional
            keywords arguments passed to the function `customize_func` at call.

        Examples
        --------
        >>> demo = load_example_data('demography_eurostat')
        >>> report = ExcelReport(EXAMPLE_EXCEL_TEMPLATES_DIR)

        >>> sheet_population = report.new_sheet('Population')
        >>> population = demo.population

        Generate a new graph for each combination of gender and year

        >>> sheet_population.add_graphs(
        ...     {'Population of {gender} by country in {year}': population},
        ...     {'gender': population.gender, 'year': population.time},
        ...     template='line', width=450, height=250, graphs_per_row=2)

        Specify the mininum and maximum values for the Y axis

        >>> sheet_population.add_graphs({'Population of {gender} by country for the year {year}': population},
        ...                      {'gender': population.gender, 'year': population.time},
        ...                      template='line', width=450, height=250, graphs_per_row=2, min_y=0, max_y=50e6)

        Specify the interval between two ticks (X axis)

        >>> sheet_population.add_graphs({'Population of {gender} by country for the year {year}': population},
        ...                      {'gender': population.gender, 'year': population.time},
        ...                      template='line', width=450, height=250, graphs_per_row=2, xticks_spacing=2)

        >>> # do not forget to call 'to_excel' to create the report file
        >>> report.to_excel('Demography_Report.xlsx')
        """
        pass

    def newline(self):
        r"""
        Force a new row of graphs.
        """
        pass


class AbstractExcelReport(AbstractReportItem):
    r"""
    Automate the generation of multiple graphs in an Excel file.

    The ExcelReport instance is initially populated with information
    (data, title, destination sheet, template, size) required to create the graphs.
    Once all information has been provided, the :py:obj:`~ExcelReport.to_excel` method
    is called to generate an Excel file with all graphs in one step.

    Parameters
    ----------
    template_dir : str, optional
        Path to the directory containing the Excel template files (with a '.crtx' extension).
        Defaults to None.
    template : str, optional
        Name of the template to be used as default template.
        The extension '.crtx' will be added if not given.
        The full path to the template file must be given if no template directory has been set.
        Defaults to None.
    graphs_per_row: int, optional
        Default number of graphs per row.
        Defaults to 1.

    Notes
    -----
    The data associated with all graphical items is dumped in the same sheet named '__data__'.

    Examples
    --------
    >>> demo = load_example_data('demography_eurostat')
    >>> report = ExcelReport(EXAMPLE_EXCEL_TEMPLATES_DIR)

    Set a new destination sheet

    >>> sheet_be = report.new_sheet('Belgium')

    Add a new title item

    >>> sheet_be.add_title('Population, births and deaths')

    Add a new graph item (each new graph is placed right to previous one unless you use newline() or add_title())

    >>> # using default 'width' and 'height' values
    >>> sheet_be.add_graph(demo.population['Belgium'], 'Population', template='Line')
    >>> # specifying the 'width' and 'height' values
    >>> sheet_be.add_graph(demo.births['Belgium'], 'Births', template='Line', width=450, height=250)

    Override the default 'width' and 'height' values for graphs

    >>> sheet_be.set_item_default_size('graph', width=450, height=250)
    >>> # add a new graph with the new default 'width' and 'height' values
    >>> sheet_be.add_graph(demo.deaths['Belgium'], 'Deaths')

    Set a default template for all next graphs

    >>> # if a default template directory has been set, just pass the name
    >>> sheet_be.template = 'Line'
    >>> # otherwise, give the full path to the template file
    >>> sheet_be.template = r'C:\other_template_dir\Line_Marker.crtx' # doctest: +SKIP
    >>> # add a new graph with the default template
    >>> sheet_be.add_graph(demo.population['Belgium', 'Female'], 'Population - Female')
    >>> sheet_be.add_graph(demo.population['Belgium', 'Male'], 'Population - Male')

    Specify the number of graphs per row

    >>> sheet_countries = report.new_sheet('All countries')

    >>> sheet_countries.graphs_per_row = 2
    >>> for combined_labels, subset in demo.population.items(('time', 'gender')):
    ...    title = ' - '.join([str(label) for label in combined_labels])
    ...    sheet_countries.add_graph(subset, title)

    Force a new row of graphs

    >>> sheet_countries.newline()

    Add multiple graphs at once (add a new graph for each combination of gender and year)

    >>> sheet_countries.add_graphs({'Population of {gender} by country in {year}': population},
    ...                            {'gender': population.gender, 'year': population.time},
    ...                            template='line', width=450, height=250, graphs_per_row=2)

    Generate the report Excel file

    >>> report.to_excel('Demography_Report.xlsx')
    """

    def new_sheet(self, sheet_name):
        r"""
        Add a new empty output sheet.
        This sheet will contain only graphical elements, all data are exported
        to a dedicated separate sheet.

        Parameters
        ----------
        sheet_name : str
            name of the current sheet.

        Returns
        -------
        sheet: ReportSheet

        Examples
        --------
        >>> demo = load_example_data('demography_eurostat')
        >>> report = ExcelReport(EXAMPLE_EXCEL_TEMPLATES_DIR)

        >>> # prepare new output sheet named 'Belgium'
        >>> sheet_be = report.new_sheet('Belgium')

        >>> # add graph to the output sheet 'Belgium'
        >>> sheet_be.add_graph(demo.population['Belgium'], 'Population', template='Line')
        """
        pass

    def sheet_names(self):
        r"""
        Returns the names of the output sheets.

        Examples
        --------
        >>> report = ExcelReport()
        >>> sheet_population = report.new_sheet('Pop')
        >>> sheet_births = report.new_sheet('Births')
        >>> sheet_deaths = report.new_sheet('Deaths')
        >>> report.sheet_names()
        ['Pop', 'Births', 'Deaths']
        """
        pass

    def to_excel(self, filepath, data_sheet_name='__data__', overwrite=True):
        r"""
        Generate the report Excel file.

        Parameters
        ----------
        filepath : str
            Path of the report file for the dump.
        data_sheet_name : str, optional
            name of the Excel sheet where all data associated with items is dumped.
            Defaults to '__data__'.
        overwrite : bool, optional
            whether or not to overwrite an existing report file.
            Defaults to True.

        Examples
        --------
        >>> demo = load_example_data('demography_eurostat')
        >>> report = ExcelReport(EXAMPLE_EXCEL_TEMPLATES_DIR)
        >>> report.template = 'Line_Marker'

        >>> for c in demo.country:
        ...     sheet_country = report.new_sheet(c)
        ...     sheet_country.add_graph(demo.population[c], 'Population')
        ...     sheet_country.add_graph(demo.births[c], 'Births')
        ...     sheet_country.add_graph(demo.deaths[c], 'Deaths')

        Basic usage

        >>> report.to_excel('Demography_Report.xlsx')

        Alternative data sheet name

        >>> report.to_excel('Demography_Report.xlsx', data_sheet_name='Data Tables') # doctest: +SKIP

        Check if ouput file already exists

        >>> report.to_excel('Demography_Report.xlsx', overwrite=False) # doctest: +SKIP
        Traceback (most recent call last):
        ...
        ValueError: Sheet named 'Belgium' already present in workbook
        """
        pass


if xw is not None:
    from xlwings.constants import LegendPosition, HAlign, VAlign, ChartType, RowCol, AxisType, Constants
    from larray.inout.xw_excel import open_excel

    class ItemSize:
        def __init__(self, width, height):
            self.width = width
            self.height = height

        @property
        def width(self):
            return self._width

        @width.setter
        def width(self, width):
            _positive_integer(width)
            self._width = width

        @property
        def height(self):
            return self._height

        @height.setter
        def height(self, height):
            _positive_integer(height)
            self._height = height


    class ExcelTitleItem(ItemSize):

        _default_size = ItemSize(1000, 50)

        def __init__(self, text, fontsize, top, left, width, height):
            ItemSize.__init__(self, width, height)
            self.top = top
            self.left = left
            self.text = str(text)
            _positive_integer(fontsize)
            self.fontsize = fontsize

        def dump(self, sheet, data_sheet, row):
            data_cells = data_sheet.Cells

            # add title in data sheet
            data_cells(row, 1).Value = self.text

            # generate title banner in destination sheet
            msoShapeRectangle = 1
            msoThemeColorBackground1 = 14
            sheet_shapes = sheet.Shapes
            shp = sheet_shapes.AddShape(Type=msoShapeRectangle, Left=self.left, Top=self.top,
                                        Width=self.width, Height=self.height)
            fill = shp.Fill
            fill.ForeColor.ObjectThemeColor = msoThemeColorBackground1
            fill.Solid()
            shp.Line.Visible = False
            frame = shp.TextFrame
            chars = frame.Characters()
            chars.Text = self.text
            font = chars.Font
            font.Color = 1
            font.Bold = True
            font.Size = self.fontsize
            frame.HorizontalAlignment = HAlign.xlHAlignLeft
            frame.VerticalAlignment = VAlign.xlVAlignCenter
            shp.SetShapesDefaultProperties()
            # update and return current row position in data sheet (+1 for title +1 for blank line)
            return row + 2

    _default_items_size['title'] = ExcelTitleItem._default_size

    class ExcelGraphItem(ItemSize):

        _default_size = ItemSize(427, 230)

        def __init__(self, data, title, template, top, left, width, height, min_y, max_y,
                     xticks_spacing, customize_func, customize_kwargs):
            ItemSize.__init__(self, width, height)
            self.top = top
            self.left = left
            self.title = str(title) if title is not None else None
            data = asarray(data)
            if not (1 <= data.ndim <= 2):
                raise ValueError(f"Expected 1D or 2D array for data argument. Got array of dimensions {data.ndim}")
            self.data = data
            if template is not None and not os.path.isfile(template):
                raise ValueError(f"Could not find template file {template}")
            self.template = template
            self.min_y = min_y
            self.max_y = max_y
            self.xticks_spacing = xticks_spacing
            if customize_func is not None and not callable(customize_func):
                raise TypeError(f"Expected a function for the argument 'customize_func'. "
                                f"Got object of type {type(customize_func).__name__} instead.")
            self.customize_func = customize_func
            self.customize_kwargs = customize_kwargs

        def dump(self, sheet, data_sheet, row):
            data_range = data_sheet.Range
            data_cells = data_sheet.Cells

            # write graph title in data sheet
            data_cells(row, 1).Value = self.title
            row += 1

            # dump data to make the graph in data sheet
            data = self.data
            nb_series = 1 if data.ndim == 1 else data.shape[0]
            nb_xticks = data.size if data.ndim == 1 else data.shape[1]
            last_row, last_col = row + nb_series, nb_xticks + 1
            data_range(data_cells(row, 1), data_cells(last_row, last_col)).Value = data.dump(na_repr=None)
            data_cells(row, 1).Value = ''

            # generate graph in destination sheet
            sheet_charts = sheet.ChartObjects()
            obj = sheet_charts.Add(self.left, self.top, self.width, self.height)
            obj_chart = obj.Chart
            source = data_range(data_cells(row, 1), data_cells(last_row, last_col))
            obj_chart.SetSourceData(source)
            obj_chart.ChartType = ChartType.xlLine
            # title
            if self.title is not None:
                obj_chart.HasTitle = True
                obj_chart.ChartTitle.Caption = self.title
            # legend
            obj_chart.Legend.Position = LegendPosition.xlLegendPositionBottom
            # template
            if self.template is not None:
                obj_chart.ApplyChartTemplate(self.template)
            # min - max on Y axis
            if self.min_y is not None:
                obj_chart.Axes(AxisType.xlValue).MinimumScale = self.min_y
            if self.max_y is not None:
                obj_chart.Axes(AxisType.xlValue).MaximumScale = self.max_y
            # xticks_spacing
            if self.xticks_spacing is not None:
                obj_chart.Axes(AxisType.xlCategory).TickLabelSpacing = self.xticks_spacing
                obj_chart.Axes(AxisType.xlCategory).TickMarkSpacing = self.xticks_spacing
                obj_chart.Axes(AxisType.xlCategory).TickLabelPosition = Constants.xlLow
            # user's function (to apply on remaining kwargs)
            if self.customize_func is not None:
                self.customize_func(obj_chart, **self.customize_kwargs)
            # flagflip
            if nb_series > 1 and nb_xticks == 1:
                obj_chart.PlotBy = RowCol.xlRows
            # update and return current row position
            return row + nb_series + 2

    _default_items_size['graph'] = ExcelGraphItem._default_size

    class ReportSheet(AbstractReportSheet):
        def __init__(self, excel_report, name, template_dir=None, template=None, graphs_per_row=1):
            name = _translate_sheet_name(name)
            self.excel_report = excel_report
            self.name = name
            self.items = []
            self.top = 0
            self.left = 0
            self.position_in_row = 1
            self.curline_height = 0
            if template_dir is None:
                template_dir = excel_report.template_dir
            if template is None:
                template = excel_report.template
            AbstractReportSheet.__init__(self, template_dir, template, graphs_per_row)

        def add_title(self, title, width=None, height=None, fontsize=11):
            if width is None:
                width = self.default_items_size['title'].width
            if height is None:
                height = self.default_items_size['title'].height
            self.newline()
            self.items.append(ExcelTitleItem(title, fontsize, self.top, 0, width, height))
            self.top += height

        def add_graph(self, data, title=None, template=None, width=None, height=None, min_y=None, max_y=None,
                      xticks_spacing=None, customize_func=None, customize_kwargs=None):
            if width is None:
                width = self.default_items_size['graph'].width
            if height is None:
                height = self.default_items_size['graph'].height
            if template is not None:
                self.template = template
            template = self.template
            if self.graphs_per_row is not None and self.position_in_row > self.graphs_per_row:
                self.newline()
            self.items.append(ExcelGraphItem(data, title, template, self.top, self.left, width, height,
                                             min_y, max_y, xticks_spacing, customize_func, customize_kwargs))
            self.left += width
            self.curline_height = max(self.curline_height, height)
            self.position_in_row += 1

        def add_graphs(self, array_per_title, axis_per_loop_variable, template=None, width=None, height=None,
                       graphs_per_row=1, min_y=None, max_y=None, xticks_spacing=None, customize_func=None,
                       customize_kwargs=None):
            loop_variable_names = axis_per_loop_variable.keys()
            axes = tuple(axis_per_loop_variable.values())
            titles = array_per_title.keys()
            arrays = array_per_title.values()
            if graphs_per_row is not None:
                previous_graphs_per_row = self.graphs_per_row
                self.graphs_per_row = graphs_per_row
            if self.position_in_row > 1:
                self.newline()
            for loop_variable_values, arrays_chunk in zip_array_items(arrays, axes=axes):
                loop_variables_dict = dict(zip(loop_variable_names, loop_variable_values))
                for title_template, array_chunk in zip(titles, arrays_chunk):
                    title = title_template.format(**loop_variables_dict)
                    self.add_graph(array_chunk, title, template, width, height, min_y, max_y, xticks_spacing,
                                   customize_func, customize_kwargs)
            if graphs_per_row is not None:
                self.graphs_per_row = previous_graphs_per_row

        def newline(self):
            self.top += self.curline_height
            self.curline_height = 0
            self.left = 0
            self.position_in_row = 1

        def _to_excel(self, workbook, data_row):
            # use first sheet as data sheet
            data_sheet = workbook.Worksheets(1)
            data_cells = data_sheet.Cells
            # write destination sheet name in data sheet
            data_cells(data_row, 1).Value = self.name
            data_row += 2

            # create new empty sheet in workbook (will contain output graphical items)
            # Hack, since just specifying "After" is broken in certain environments
            # see: https://stackoverflow.com/questions/40179804/adding-excel-sheets-to-end-of-workbook
            dest_sheet = workbook.Worksheets.Add(Before=None, After=workbook.Sheets(workbook.Sheets.Count))
            dest_sheet.Name = self.name
            # for each item, dump data + generate associated graphical items
            for item in self.items:
                data_row = item.dump(dest_sheet, data_sheet, data_row)
            # reset
            self.top = 0
            self.left = 0
            self.curline_height = 0
            # return current row in data sheet
            return data_row

    # TODO : add a new section about this class in the tutorial
    class ExcelReport(AbstractExcelReport):
        def __init__(self, template_dir=None, template=None, graphs_per_row=1):
            AbstractExcelReport.__init__(self, template_dir, template, graphs_per_row)
            self.sheets = OrderedDict()

        def sheet_names(self):
            return [sheet_name for sheet_name in self.sheets.keys()]

        def __getitem__(self, key):
            return self.sheets[key]

        # TODO : Do not implement __setitem__ and move code below to new_sheet()?
        def __setitem__(self, key, value, warn_stacklevel=2):
            if not isinstance(value, ReportSheet):
                raise ValueError(f"Expected ReportSheet object. Got {type(value).__name__} object instead.")
            if key in self.sheet_names():
                warnings.warn(f"Sheet '{key}' already exists in the report and will be reset",
                              stacklevel=warn_stacklevel)
            self.sheets[key] = value

        def __delitem__(self, key):
            del self.sheets[key]

        def __repr__(self):
            return f'sheets: {self.sheet_names()}'

        def new_sheet(self, sheet_name):
            sheet = ReportSheet(self, sheet_name, self.template_dir, self.template, self.graphs_per_row)
            self.__setitem__(sheet_name, sheet, warn_stacklevel=3)
            return sheet

        def to_excel(self, filepath, data_sheet_name='__data__', overwrite=True):
            with open_excel(filepath, overwrite_file=overwrite) as wb:
                # from here on, we use pure win32com objects instead of
                # larray.excel or xlwings objects as this is faster
                xl_wb = wb.api

                # rename first sheet
                xl_wb.Worksheets(1).Name = data_sheet_name

                # dump items for each output sheet
                data_sheet_row = 1
                for sheet in self.sheets.values():
                    data_sheet_row = sheet._to_excel(xl_wb, data_sheet_row)
                wb.save()
                # reset
                self.sheets.clear()
else:
    class ReportSheet(AbstractReportSheet):
        def __init__(self):
            raise Exception("ReportSheet class cannot be instantiated because xlwings is not installed")

    class ExcelReport(AbstractExcelReport):
        def __init__(self):
            raise Exception("ExcelReport class cannot be instantiated because xlwings is not installed")


ExcelReport.__doc__ = AbstractExcelReport.__doc__
ReportSheet.__doc__ = AbstractReportSheet.__doc__



from collections import OrderedDict


class AttributeDict(OrderedDict):
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        del self[key]

    def __dir__(self):
        return list(set(super(AttributeDict, self).__dir__()) | set(self.keys()))

    def __repr__(self):
        return '\n'.join([f'{k}: {v}' for k, v in self.items()])


class Metadata(AttributeDict):
    r"""
    An ordered dictionary allowing key-values accessibly using attribute notation (AttributeDict.attribute)
    instead of key notation (Dict["key"]).

    Examples
    --------
    >>> from larray import ndtest
    >>> from datetime import datetime

    Add metadata at array initialization

    >>> arr = ndtest((3, 3), meta=Metadata(title='the title', author='John Smith'))

    Add metadata after array initialization

    >>> arr.meta.creation_date = datetime(2017, 2, 10)

    Access to metadata

    >>> arr.meta.creation_date
    datetime.datetime(2017, 2, 10, 0, 0)

    Modify metadata

    >>> arr.meta.creation_date = datetime(2017, 2, 16)

    Delete metadata

    >>> del arr.meta.creation_date
    """
    def __larray__(self):
        from larray.core.array import stack
        return stack(self.items(), axes='metadata')

    @classmethod
    def from_array(cls, array):
        from larray.core.array import asarray
        array = asarray(array)
        if array.ndim != 1:
            raise ValueError(f"Expected Array object of dimension 1. Got array of dimension {array.ndim}")

        from pandas import to_numeric, to_datetime

        def _convert_value(value):
            value = to_numeric([value], errors='ignore')[0]
            if isinstance(value, str):
                value = to_datetime(value, errors='ignore', infer_datetime_format=True)
            return value

        return Metadata([(key, _convert_value(value)) for key, value in zip(array.axes.labels[0], array.data)])

    # ---------- IO methods ----------
    def to_hdf(self, hdfstore, key=None):
        if len(self):
            attrs = hdfstore.get_storer(key).attrs if key is not None else hdfstore.root._v_attrs
            attrs.metadata = self

    @classmethod
    def from_hdf(cls, hdfstore, key=None):
        attrs = hdfstore.get_storer(key).attrs if key is not None else hdfstore.root._v_attrs
        if 'metadata' in attrs:
            return attrs.metadata
        else:
            return None



import os
import larray as la


_TEST_DIR = os.path.join(os.path.dirname(__file__), 'tests')

EXAMPLE_FILES_DIR = os.path.join(_TEST_DIR, 'data')
AVAILABLE_EXAMPLE_DATA = {
    'demography': os.path.join(EXAMPLE_FILES_DIR, 'demography.h5'),
    'demography_eurostat': os.path.join(EXAMPLE_FILES_DIR, 'demography_eurostat.h5')
}
EXAMPLE_EXCEL_TEMPLATES_DIR = os.path.join(_TEST_DIR, 'excel_template')


def get_example_filepath(fname):
    r"""Return absolute path to an example file if exist.

    Parameters
    ----------
    fname : str
        Filename of an existing example file.

    Returns
    -------
    Filepath
        Absolute filepath to an example file if exists.

    Notes
    -----
    A ValueError is raised if the provided filename does not represent an existing example file.

    Examples
    --------
    >>> fpath = get_example_filepath('examples.xlsx')
    """
    fpath = os.path.abspath(os.path.join(EXAMPLE_FILES_DIR, fname))
    if not os.path.exists(fpath):
        AVAILABLE_EXAMPLE_FILES = os.listdir(EXAMPLE_FILES_DIR)
        raise ValueError(f"Example file {fname} does not exist. "
                         f"Available example files are: {AVAILABLE_EXAMPLE_FILES}")
    return fpath


# Note that we skip doctests because they require pytables, which is only an optional dependency and its hard
# to skip doctests selectively.
# CHECK: We might want to use .csv files for the example data, so that it can be loaded with any optional dependency.
def load_example_data(name):
    r"""Load arrays used in the tutorial so that all examples in it can be reproduced.

    Parameters
    ----------
    name : str
        Example data to load. Available example datasets are:

        - demography
        - demography_eurostat

    Returns
    -------
    Session
        Session containing one or several arrays.

    Examples
    --------
    >>> demo = load_example_data('demography')           # doctest: +SKIP
    >>> print(demo.summary())                            # doctest: +SKIP
    hh: time, geo, hh_type (26 x 3 x 7) [int64]
    pop: time, geo, age, sex, nat (26 x 3 x 121 x 2 x 2) [int64]
    qx: time, geo, age, sex, nat (26 x 3 x 121 x 2 x 2) [float64]
    >>> demo = load_example_data('demography_eurostat')  # doctest: +SKIP
    >>> print(demo.summary())                            # doctest: +SKIP
    Metadata:
       title: Demographic datasets for a small selection of countries in Europe
       source: demo_jpan, demo_fasec, demo_magec and migr_imm1ctz tables from Eurostat
    gender: gender ['Male' 'Female'] (2)
    country: country ['Belgium' 'France' 'Germany'] (3)
    country_benelux: country_benelux ['Belgium' 'Luxembourg' 'Netherlands'] (3)
    citizenship: citizenship ['Belgium' 'Luxembourg' 'Netherlands'] (3)
    time: time [2013 2014 2015 2016 2017] (5)
    even_years: time[2014 2016] >> even_years (2)
    odd_years: time[2013 2015 2017] >> odd_years (3)
    births: country, gender, time (3 x 2 x 5) [int32]
    deaths: country, gender, time (3 x 2 x 5) [int32]
    immigration: country, citizenship, gender, time (3 x 3 x 2 x 5) [int32]
    pop: country, gender, time (3 x 2 x 5) [int32]
    pop_benelux: country, gender, time (3 x 2 x 5) [int32]
    """
    if name is None:
        name = 'demography'
    if not isinstance(name, str):
        raise TypeError("Expected string for argument example_data")
    if name not in AVAILABLE_EXAMPLE_DATA:
        available_datasets = list(AVAILABLE_EXAMPLE_DATA.keys())
        raise ValueError(f"example_data must be chosen from list {available_datasets}")
    return la.Session(AVAILABLE_EXAMPLE_DATA[name])



# copy-pasted from SQLAlchemy util/_collections.py

# Copyright (C) 2005-2015 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

from larray.util.misc import unique_list


class OrderedSet(set):
    def __init__(self, d=None):
        set.__init__(self)
        if d is not None:
            self._list = unique_list(d)
            set.update(self, self._list)
        else:
            self._list = []

    def add(self, element):
        if element not in self:
            self._list.append(element)
        set.add(self, element)

    def remove(self, element):
        set.remove(self, element)
        self._list.remove(element)

    def insert(self, pos, element):
        if element not in self:
            self._list.insert(pos, element)
        set.add(self, element)

    def discard(self, element):
        if element in self:
            self._list.remove(element)
            set.remove(self, element)

    def clear(self):
        set.clear(self)
        self._list = []

    def __getitem__(self, key):
        return self._list[key]

    def __iter__(self):
        return iter(self._list)

    def __repr__(self):
        class_name = self.__class__.__name__
        return f'{class_name}({self._list!r})'

    __str__ = __repr__

    def update(self, iterable):
        for e in iterable:
            if e not in self:
                self._list.append(e)
                set.add(self, e)
        return self
    __ior__ = update

    def union(self, other):
        result = self.__class__(self)
        result.update(other)
        return result
    __or__ = union
    __add__ = union

    def intersection(self, other):
        other = set(other)
        return self.__class__(a for a in self if a in other)
    __and__ = intersection

    def symmetric_difference(self, other):
        other = set(other)
        result = self.__class__(a for a in self if a not in other)
        result.update(a for a in other if a not in self)
        return result
    __xor__ = symmetric_difference

    def difference(self, other):
        other = set(other)
        return self.__class__(a for a in self if a not in other)
    __sub__ = difference

    def intersection_update(self, other):
        other = set(other)
        set.intersection_update(self, other)
        self._list = [a for a in self._list if a in other]
        return self
    __iand__ = intersection_update

    def symmetric_difference_update(self, other):
        set.symmetric_difference_update(self, other)
        self._list = [a for a in self._list if a in self]
        self._list += [a for a in other._list if a in self]
        return self
    __ixor__ = symmetric_difference_update

    def difference_update(self, other):
        set.difference_update(self, other)
        self._list = [a for a in self._list if a in self]
        return self
    __isub__ = difference_update



from abc import ABCMeta
from copy import deepcopy
import warnings

import numpy as np

from typing import TYPE_CHECKING, Type, Any, Dict, Set, List, no_type_check

from larray.core.metadata import Metadata
from larray.core.axis import AxisCollection
from larray.core.group import Group
from larray.core.array import Array, full
from larray.core.session import Session


class NotLoaded:
    pass


try:
    import pydantic
except ImportError:
    pydantic = None

#  moved the not implemented versions of Checked* classes in the beginning of the module
#  otherwise PyCharm do not provide auto-completion for methods of CheckedSession
#  (imported from Session)
if not pydantic:
    def CheckedArray(axes: AxisCollection, dtype: np.dtype = float) -> Type[Array]:
        raise NotImplementedError("CheckedArray cannot be used because pydantic is not installed")

    class CheckedSession:
        def __init__(self, *args, **kwargs):
            raise NotImplementedError("CheckedSession class cannot be instantiated "
                                      "because pydantic is not installed")

    class CheckedParameters:
        def __init__(self, *args, **kwargs):
            raise NotImplementedError("CheckedParameters class cannot be instantiated "
                                      "because pydantic is not installed")
else:
    from pydantic.fields import ModelField
    from pydantic.class_validators import Validator
    from pydantic.main import BaseConfig

    # the implementation of the class below is inspired by the 'ConstrainedBytes' class
    # from the types.py module of the 'pydantic' library
    class CheckedArrayImpl(Array):
        expected_axes: AxisCollection
        dtype: np.dtype = np.dtype(float)

        # see https://pydantic-docs.helpmanual.io/usage/types/#classes-with-__get_validators__
        @classmethod
        def __get_validators__(cls):
            # one or more validators may be yielded which will be called in the
            # order to validate the input, each validator will receive as an input
            # the value returned from the previous validator
            yield cls.validate

        @classmethod
        def validate(cls, value, field: ModelField):
            if not (isinstance(value, Array) or np.isscalar(value)):
                raise TypeError(f"Expected object of type '{Array.__name__}' or a scalar for "
                                f"the variable '{field.name}' but got object of type '{type(value).__name__}'")

            # check axes
            if isinstance(value, Array):
                error_msg = f"Array '{field.name}' was declared with axes {cls.expected_axes} but got array " \
                            f"with axes {value.axes}"
                # check for extra axes
                extra_axes = value.axes - cls.expected_axes
                if extra_axes:
                    raise ValueError(f"{error_msg} (unexpected {extra_axes} "
                                     f"{'axes' if len(extra_axes) > 1 else 'axis'})")
                # check compatible axes
                try:
                    cls.expected_axes.check_compatible(value.axes)
                except ValueError as error:
                    error_msg = str(error).replace("incompatible axes", f"Incompatible axis for array '{field.name}'")
                    raise ValueError(error_msg)
                # broadcast + transpose if needed
                value = value.expand(cls.expected_axes)
                # check dtype
                if value.dtype != cls.dtype:
                    value = value.astype(cls.dtype)
                return value
            else:
                return full(axes=cls.expected_axes, fill_value=value, dtype=cls.dtype)

    # the implementation of the function below is inspired by the 'conbytes' function
    # from the types.py module of the 'pydantic' library

    def CheckedArray(axes: AxisCollection, dtype: np.dtype = float) -> Type[Array]:
        # XXX: for a very weird reason I don't know, I have to put the fake import below
        #      to get autocompletion from PyCharm
        from larray.core.checked import CheckedArrayImpl
        """
        Represents a constrained array. It is intended to only be used along with :py:class:`CheckedSession`.

        Its axes are assumed to be "frozen", meaning they are constant all along the execution of the program.
        A constraint on the dtype of the data can be also specified.

        Parameters
        ----------
        axes: AxisCollection
            Axes of the checked array.
        dtype: data-type, optional
            Data-type for the checked array. Defaults to float.

        Returns
        -------
        Array
            Constrained array.
        """
        if axes is not None and not isinstance(axes, AxisCollection):
            axes = AxisCollection(axes)
        _dtype = np.dtype(dtype)

        class ArrayDefValue(CheckedArrayImpl):
            expected_axes = axes
            dtype = _dtype

        return ArrayDefValue

    class AbstractCheckedSession:
        pass

    # Simplified version of the ModelMetaclass class from pydantic:
    # https://github.com/samuelcolvin/pydantic/blob/master/pydantic/main.py#L195

    class ModelMetaclass(ABCMeta):
        @no_type_check  # noqa C901
        def __new__(mcs, name, bases, namespace, **kwargs):
            from pydantic.fields import Undefined
            from pydantic.class_validators import extract_validators, inherit_validators
            from pydantic.types import PyObject
            from pydantic.typing import is_classvar, resolve_annotations
            from pydantic.utils import lenient_issubclass, validate_field_name
            from pydantic.main import inherit_config, prepare_config, UNTOUCHED_TYPES

            fields: Dict[str, ModelField] = {}
            config = BaseConfig
            validators: Dict[str, List[Validator]] = {}

            for base in reversed(bases):
                if issubclass(base, AbstractCheckedSession) and base != AbstractCheckedSession:
                    config = inherit_config(base.__config__, config)
                    fields.update(deepcopy(base.__fields__))
                    validators = inherit_validators(base.__validators__, validators)

            config = inherit_config(namespace.get('Config'), config)
            validators = inherit_validators(extract_validators(namespace), validators)

            # update fields inherited from base classes
            for field in fields.values():
                field.set_config(config)
                extra_validators = validators.get(field.name, [])
                if extra_validators:
                    field.class_validators.update(extra_validators)
                    # re-run prepare to add extra validators
                    field.populate_validators()

            prepare_config(config, name)

            # extract and build fields
            class_vars = set()
            if (namespace.get('__module__'), namespace.get('__qualname__')) != \
                    ('larray.core.checked', 'CheckedSession'):
                untouched_types = UNTOUCHED_TYPES + config.keep_untouched

                # annotation only fields need to come first in fields
                annotations = resolve_annotations(namespace.get('__annotations__', {}),
                                                  namespace.get('__module__', None))
                for ann_name, ann_type in annotations.items():
                    if is_classvar(ann_type):
                        class_vars.add(ann_name)
                    elif not ann_name.startswith('_'):
                        validate_field_name(bases, ann_name)
                        value = namespace.get(ann_name, Undefined)
                        if (isinstance(value, untouched_types) and ann_type != PyObject
                                and not lenient_issubclass(getattr(ann_type, '__origin__', None), Type)):
                            continue
                        fields[ann_name] = ModelField.infer(name=ann_name, value=value, annotation=ann_type,
                                                            class_validators=validators.get(ann_name, []),
                                                            config=config)

                for var_name, value in namespace.items():
                    # 'var_name not in annotations' because namespace.items() contains annotated fields
                    # with default values
                    # 'var_name not in class_vars' to avoid to update a field if it was redeclared (by mistake)
                    if (var_name not in annotations and not var_name.startswith('_')
                            and not isinstance(value, untouched_types) and var_name not in class_vars):
                        validate_field_name(bases, var_name)
                        # the method ModelField.infer() fails to infer the type of Group objects
                        # (which are interpreted as ndarray objects)
                        annotation = type(value) if isinstance(value, Group) else annotations.get(var_name)
                        inferred = ModelField.infer(name=var_name, value=value, annotation=annotation,
                                                    class_validators=validators.get(var_name, []), config=config)
                        if var_name in fields and inferred.type_ != fields[var_name].type_:
                            raise TypeError(f'The type of {name}.{var_name} differs from the new default value; '
                                            f'if you wish to change the type of this field, please use a type '
                                            f'annotation')
                        fields[var_name] = inferred

            new_namespace = {
                '__config__': config,
                '__fields__': fields,
                '__field_defaults__': {n: f.default for n, f in fields.items() if not f.required},
                '__validators__': validators,
                **{n: v for n, v in namespace.items() if n not in fields},
            }
            return super().__new__(mcs, name, bases, new_namespace, **kwargs)

    class CheckedSession(Session, AbstractCheckedSession, metaclass=ModelMetaclass):
        """
        This class is intended to be inherited by user defined classes in which the variables of a model are declared.
        Each declared variable is constrained by a type defined explicitly or deduced from the given default value
        (see examples below).
        All classes inheriting from `CheckedSession` will have access to all methods of the :py:class:`Session` class.

        The special :py:funct:`ConsArray` type represents an Array object with fixed axes and/or dtype.
        This prevents users from modifying the dimensions (and labels) and/or the dtype of an array by mistake
        and make sure that the definition of an array remains always valid in the model.

        By declaring variables, users will speed up the development of their models using the auto-completion
        (the feature in which development tools like PyCharm try to predict the variable or function a user intends
        to enter after only a few characters have been typed).

        As for normal Session objects, it is still possible to add undeclared variables to instances of
        classes inheriting from `CheckedSession` but this must be done with caution.

        Parameters
        ----------
        *args : str or dict of {str: object} or iterable of tuples (str, object)
            Path to the file containing the session to load or
            list/tuple/dictionary containing couples (name, object).
        **kwargs : dict of {str: object}

            * Objects to add written as name=object
            * meta : list of pairs or dict or OrderedDict or Metadata, optional
                Metadata (title, description, author, creation_date, ...) associated with the array.
                Keys must be strings. Values must be of type string, int, float, date, time or datetime.

        Warnings
        --------
        The :py:method:`CheckedSession.filter`, :py:method:`CheckedSession.compact`
        and :py:method:`CheckedSession.apply` methods return a simple Session object.
        The type of the declared variables (and the value for the declared constants) will
        no longer be checked.

        See Also
        --------
        Session, CheckedParameters

        Examples
        --------

        Content of file 'parameters.py'

        >>> from larray import *
        >>> FIRST_YEAR = 2020
        >>> LAST_YEAR = 2030
        >>> AGE = Axis('age=0..10')
        >>> GENDER = Axis('gender=male,female')
        >>> TIME = Axis(f'time={FIRST_YEAR}..{LAST_YEAR}')

        Content of file 'model.py'

        >>> class ModelVariables(CheckedSession):
        ...     # --- declare variables with defined types ---
        ...     # Their values will be defined at runtime but must match the specified type.
        ...     birth_rate: Array
        ...     births: Array
        ...     # --- declare variables with a default value ---
        ...     # The default value will be used to set the variable if no value is passed at instantiation (see below).
        ...     # Their type is deduced from their default value and cannot be changed at runtime.
        ...     target_age = AGE[:2] >> '0-2'
        ...     population = zeros((AGE, GENDER, TIME), dtype=int)
        ...     # --- declare checked arrays ---
        ...     # The checked arrays have axes assumed to be "frozen", meaning they are
        ...     # constant all along the execution of the program.
        ...     mortality_rate: CheckedArray((AGE, GENDER))
        ...     # For checked arrays, the default value can be given as a scalar.
        ...     # Optionally, a dtype can be also specified (defaults to float).
        ...     deaths: CheckedArray((AGE, GENDER, TIME), dtype=int) = 0

        >>> variant_name = "baseline"
        >>> # Instantiation --> create an instance of the ModelVariables class.
        >>> # Warning: All variables declared without a default value must be set.
        >>> m = ModelVariables(birth_rate = zeros((AGE, GENDER)),
        ...                    births = zeros((AGE, GENDER, TIME), dtype=int),
        ...                    mortality_rate = 0)

        >>> # ==== model ====
        >>> # In the definition of ModelVariables, the 'birth_rate' variable, has been declared as an Array object.
        >>> # This means that the 'birth_rate' variable will always remain of type Array.
        >>> # Any attempt to assign a non-Array value to 'birth_rate' will make the program to crash.
        >>> m.birth_rate = Array([0.045, 0.055], GENDER)    # OK
        >>> m.birth_rate = [0.045, 0.055]                   # Fails
        Traceback (most recent call last):
            ...
        pydantic.errors.ArbitraryTypeError: instance of Array expected
        >>> # However, the arrays 'birth_rate', 'births' and 'population' have not been declared as 'CheckedArray'.
        >>> # Thus, axes and dtype of these arrays are not protected, leading to potentially unexpected behavior
        >>> # of the model.
        >>> # example 1: Let's say we want to calculate the new births for the year 2025 and we assume that
        >>> # the birth rate only differ by gender.
        >>> # In the line below, we add an additional TIME axis to 'birth_rate' while it was initialized
        >>> # with the AGE and GENDER axes only
        >>> m.birth_rate = full((AGE, GENDER, TIME), fill_value=Array([0.045, 0.055], GENDER))
        >>> # here 'new_births' have the AGE, GENDER and TIME axes instead of the AGE and GENDER axes only
        >>> new_births = m.population['female', 2025] * m.birth_rate
        >>> print(new_births.info)
        11 x 2 x 11
         age [11]: 0 1 2 ... 8 9 10
         gender [2]: 'male' 'female'
         time [11]: 2020 2021 2022 ... 2028 2029 2030
        dtype: float64
        memory used: 1.89 Kb
        >>> # and the line below will crash
        >>> m.births[2025] = new_births         # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Value {time} axis is not present in target subset {age, gender}.
        A value can only have the same axes or fewer axes than the subset being targeted
        >>> # now let's try to do the same for deaths and making the same mistake as for 'birth_rate'.
        >>> # The program will crash now at the first step instead of letting you go further
        >>> m.mortality_rate = full((AGE, GENDER, TIME), fill_value=sequence(AGE, inc=0.02))    \
        # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Array 'mortality_rate' was declared with axes {age, gender} but got array with axes
        {age, gender, time} (unexpected {time} axis)

        >>> # example 2: let's say we want to calculate the new births for all years.
        >>> m.birth_rate = full((AGE, GENDER, TIME), fill_value=Array([0.045, 0.055], GENDER))
        >>> new_births = m.population['female'] * m.birth_rate
        >>> # here 'new_births' has the same axes as 'births' but is a float array instead of
        >>> # an integer array as 'births'.
        >>> # The line below will make the 'births' array become a float array while
        >>> # it was initialized as an integer array
        >>> m.births = new_births
        >>> print(m.births.info)
        11 x 11 x 2
         age [11]: 0 1 2 ... 8 9 10
         time [11]: 2020 2021 2022 ... 2028 2029 2030
         gender [2]: 'male' 'female'
        dtype: float64
        memory used: 1.89 Kb
        >>> # now let's try to do the same for deaths.
        >>> m.mortality_rate = full((AGE, GENDER), fill_value=sequence(AGE, inc=0.02))
        >>> # here the result of the multiplication of the 'population' array by the 'mortality_rate' array
        >>> # is automatically converted to an integer array
        >>> m.deaths = m.population * m.mortality_rate
        >>> print(m.deaths.info)                            # doctest: +SKIP
        11 x 2 x 11
         age [11]: 0 1 2 ... 8 9 10
         gender [2]: 'male' 'female'
         time [11]: 2020 2021 2022 ... 2028 2029 2030
        dtype: int32
        memory used: 968 bytes

        >>> # note that it still possible to add undeclared variables to a checked session
        >>> # but this must be done with caution.
        >>> m.undeclared_var = 'undeclared_var'

        >>> # ==== output ====
        >>> # save all variables in an HDF5 file
        >>> m.save(f'{variant_name}.h5', display=True)
        dumping birth_rate ... done
        dumping births ... done
        dumping mortality_rate ... done
        dumping deaths ... done
        dumping target_age ... done
        dumping population ... done
        dumping undeclared_var ... done
        """
        if TYPE_CHECKING:
            # populated by the metaclass, defined here to help IDEs only
            __fields__: Dict[str, ModelField] = {}
            __field_defaults__: Dict[str, Any] = {}
            __validators__: Dict[str, List[Validator]] = {}
            __config__: Type[BaseConfig] = BaseConfig

        class Config:
            # whether to allow arbitrary user types for fields (they are validated simply by checking
            # if the value is an instance of the type). If False, RuntimeError will be raised on model declaration.
            # (default: False)
            arbitrary_types_allowed = True
            # whether to validate field defaults
            validate_all = True
            # whether to ignore, allow, or forbid extra attributes during model initialization (and after).
            # Accepts the string values of 'ignore', 'allow', or 'forbid', or values of the Extra enum
            # (default: Extra.ignore)
            extra = 'allow'
            # whether to perform validation on assignment to attributes
            validate_assignment = True
            # whether or not models are faux-immutable, i.e. whether __setattr__ is allowed.
            # (default: True)
            allow_mutation = True

        # Warning: order of fields is not preserved.
        # As of v1.0 of pydantic all fields with annotations (whether annotation-only or with a default value)
        # will precede all fields without an annotation. Within their respective groups, fields remain in the
        # order they were defined.
        # See https://pydantic-docs.helpmanual.io/usage/models/#field-ordering
        def __init__(self, *args, **kwargs):
            meta = kwargs.pop('meta', Metadata())
            Session.__init__(self, meta=meta)

            # create an intermediate Session object to not call the __setattr__
            # and __setitem__ overridden in the present class and in case a filepath
            # is given as only argument
            # todo: refactor Session.load() to use a private function which returns the handler directly
            # so that we can get the items out of it and avoid this
            input_data = dict(Session(*args, **kwargs))

            # --- declared variables
            for name, field in self.__fields__.items():
                value = input_data.pop(field.name, NotLoaded())

                if isinstance(value, NotLoaded):
                    if field.default is None:
                        warnings.warn(f"No value passed for the declared variable '{field.name}'", stacklevel=2)
                        self.__setattr__(name, value, skip_allow_mutation=True, skip_validation=True)
                    else:
                        self.__setattr__(name, field.default, skip_allow_mutation=True)
                else:
                    self.__setattr__(name, value, skip_allow_mutation=True)

            # --- undeclared variables
            for name, value in input_data.items():
                self.__setattr__(name, value, skip_allow_mutation=True)

        # code of the method below has been partly borrowed from pydantic.BaseModel.__setattr__()
        def _check_key_value(self, name: str, value: Any, skip_allow_mutation: bool, skip_validation: bool) -> Any:
            config = self.__config__
            if not config.extra and name not in self.__fields__:
                raise ValueError(f"Variable '{name}' is not declared in '{self.__class__.__name__}'. "
                                 f"Adding undeclared variables is forbidden. "
                                 f"List of declared variables is: {list(self.__fields__.keys())}.")
            if not skip_allow_mutation and not config.allow_mutation:
                raise TypeError(f"Cannot change the value of the variable '{name}' since '{self.__class__.__name__}' "
                                f"is immutable and does not support item assignment")
            known_field = self.__fields__.get(name, None)
            if known_field:
                if not skip_validation:
                    value, error_ = known_field.validate(value, self.dict(exclude={name}), loc=name, cls=self.__class__)
                    if error_:
                        raise error_.exc
            else:
                warnings.warn(f"'{name}' is not declared in '{self.__class__.__name__}'", stacklevel=3)
            return value

        def __setitem__(self, key, value, skip_allow_mutation=False, skip_validation=False):
            if key != 'meta':
                value = self._check_key_value(key, value, skip_allow_mutation, skip_validation)
                # we need to keep the attribute in sync
                object.__setattr__(self, key, value)
                self._objects[key] = value

        def __setattr__(self, key, value, skip_allow_mutation=False, skip_validation=False):
            if key != 'meta':
                value = self._check_key_value(key, value, skip_allow_mutation, skip_validation)
                # we need to keep the attribute in sync
                object.__setattr__(self, key, value)
            Session.__setattr__(self, key, value)

        def __getstate__(self) -> Dict[str, Any]:
            return {'__dict__': self.__dict__}

        def __setstate__(self, state: Dict[str, Any]) -> None:
            object.__setattr__(self, '__dict__', state['__dict__'])

        def dict(self, exclude: Set[str] = None):
            d = dict(self.items())
            for name in exclude:
                if name in d:
                    del d[name]
            return d

    class CheckedParameters(CheckedSession):
        """
        Same as py:class:`CheckedSession` but declared variables cannot be modified after initialization.

        Parameters
        ----------
        *args : str or dict of {str: object} or iterable of tuples (str, object)
            Path to the file containing the session to load or
            list/tuple/dictionary containing couples (name, object).
        **kwargs : dict of {str: object}

            * Objects to add written as name=object
            * meta : list of pairs or dict or OrderedDict or Metadata, optional
                Metadata (title, description, author, creation_date, ...) associated with the array.
                Keys must be strings. Values must be of type string, int, float, date, time or datetime.

        See Also
        --------
        CheckedSession

        Examples
        --------

        Content of file 'parameters.py'

        >>> from larray import *
        >>> class Parameters(CheckedParameters):
        ...     # --- declare variables with fixed values ---
        ...     # The given values can never be changed
        ...     FIRST_YEAR = 2020
        ...     LAST_YEAR = 2030
        ...     AGE = Axis('age=0..10')
        ...     GENDER = Axis('gender=male,female')
        ...     TIME = Axis(f'time={FIRST_YEAR}..{LAST_YEAR}')
        ...     # --- declare variables with defined types ---
        ...     # Their values must be defined at initialized and will be frozen after.
        ...     variant_name: str

        Content of file 'model.py'

        >>> # instantiation --> create an instance of the ModelVariables class
        >>> # all variables declared without value must be set
        >>> P = Parameters(variant_name='variant_1')
        >>> # once an instance is created, its variables can be accessed but not modified
        >>> P.variant_name
        'variant_1'
        >>> P.variant_name = 'new_variant'      # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        TypeError: Cannot change the value of the variable 'variant_name' since 'Parameters'
        is immutable and does not support item assignment
        """
        class Config:
            # whether or not models are faux-immutable, i.e. whether __setattr__ is allowed.
            # (default: True)
            allow_mutation = False



import pytest

try:
    import pydantic         # noqa: F401
except ImportError:
    pytest.skip("pydantic is required for testing Checked* classes", allow_module_level=True)

import pickle
import numpy as np

from larray import (CheckedSession, CheckedArray, Axis, AxisCollection, Group, Array,
                    ndtest, full, full_like, zeros_like, ones, ones_like, isnan)
from larray.tests.common import (inputpath, tmp_path, assert_array_nan_equal, meta,                 # noqa: F401
                                 needs_pytables, needs_openpyxl, needs_xlwings,
                                 must_warn, must_raise)
from larray.tests.test_session import (a, a2, a3, anonymous, a01, ano01, b, b2, b024,               # noqa: F401
                                       c, d, e, f, g, h,
                                       assert_seq_equal, session, test_getitem, test_getattr,
                                       test_add, test_element_equals, test_eq, test_ne)
from larray.core.checked import NotLoaded


# avoid flake8 errors
meta = meta


class TestCheckedSession(CheckedSession):
    b = b
    b024 = b024
    a: Axis
    a2: Axis
    anonymous = anonymous
    a01: Group
    ano01 = ano01
    c: str = c
    d = d
    e: Array
    g: Array
    f: CheckedArray((Axis(3), Axis(2)))
    h: CheckedArray((a3, b2), dtype=int)


@pytest.fixture()
def checkedsession():
    return TestCheckedSession(a=a, a2=a2, a01=a01, e=e, g=g, f=f, h=h)


def test_create_checkedsession_instance(meta):
    # As of v1.0 of pydantic all fields with annotations (whether annotation-only or with a default value)
    # will precede all fields without an annotation. Within their respective groups, fields remain in the
    # order they were defined.
    # See https://pydantic-docs.helpmanual.io/usage/models/#field-ordering
    declared_variable_keys = ['a', 'a2', 'a01', 'c', 'e', 'g', 'f', 'h', 'b', 'b024', 'anonymous', 'ano01', 'd']

    # setting variables without default values
    cs = TestCheckedSession(a, a01, a2=a2, e=e, f=f, g=g, h=h)
    assert list(cs.keys()) == declared_variable_keys
    assert cs.b.equals(b)
    assert cs.b024.equals(b024)
    assert cs.a.equals(a)
    assert cs.a2.equals(a2)
    assert cs.anonymous.equals(anonymous)
    assert cs.a01.equals(a01)
    assert cs.ano01.equals(ano01)
    assert cs.c == c
    assert cs.d == d
    assert cs.e.equals(e)
    assert cs.g.equals(g)
    assert cs.f.equals(f)
    assert cs.h.equals(h)

    # metadata
    cs = TestCheckedSession(a, a01, a2=a2, e=e, f=f, g=g, h=h, meta=meta)
    assert cs.meta == meta

    # override default value
    b_alt = Axis('b=b0..b4')
    cs = TestCheckedSession(a, a01, b=b_alt, a2=a2, e=e, f=f, g=g, h=h)
    assert cs.b is b_alt

    # test for "NOT_LOADED" variables
    with must_warn(UserWarning, msg="No value passed for the declared variable 'a'", check_file=False):
        TestCheckedSession(a01=a01, a2=a2, e=e, f=f, g=g, h=h)
    cs = TestCheckedSession()
    assert list(cs.keys()) == declared_variable_keys
    # --- variables with default values ---
    assert cs.b.equals(b)
    assert cs.b024.equals(b024)
    assert cs.anonymous.equals(anonymous)
    assert cs.ano01.equals(ano01)
    assert cs.c == c
    assert cs.d == d
    # --- variables without default values ---
    assert isinstance(cs.a, NotLoaded)
    assert isinstance(cs.a2, NotLoaded)
    assert isinstance(cs.a01, NotLoaded)
    assert isinstance(cs.e, NotLoaded)
    assert isinstance(cs.g, NotLoaded)
    assert isinstance(cs.f, NotLoaded)
    assert isinstance(cs.h, NotLoaded)

    # passing a scalar to set all elements a CheckedArray
    cs = TestCheckedSession(a, a01, a2=a2, e=e, f=f, g=g, h=5)
    assert cs.h.axes == AxisCollection((a3, b2))
    assert cs.h.equals(full(axes=(a3, b2), fill_value=5))

    # add the undeclared variable 'i'
    with must_warn(UserWarning, f"'i' is not declared in '{cs.__class__.__name__}'", check_file=False):
        cs = TestCheckedSession(a, a01, a2=a2, i=5, e=e, f=f, g=g, h=h)
    assert list(cs.keys()) == declared_variable_keys + ['i']

    # test inheritance between checked sessions
    class TestInheritance(TestCheckedSession):
        # override variables
        b = b2
        c: int = 5
        f: CheckedArray((a3, b2), dtype=int)
        h: CheckedArray((Axis(3), Axis(2)))
        # new variables
        n0 = 'first new var'
        n1: str

    declared_variable_keys += ['n1', 'n0']
    cs = TestInheritance(a, a01, a2=a2, e=e, f=h, g=g, h=f, n1='second new var')
    assert list(cs.keys()) == declared_variable_keys
    # --- overriden variables ---
    assert cs.b.equals(b2)
    assert cs.c == 5
    assert cs.f.equals(h)
    assert cs.h.equals(f)
    # --- new variables ---
    assert cs.n0 == 'first new var'
    assert cs.n1 == 'second new var'
    # --- variables declared in the base class ---
    assert cs.b024.equals(b024)
    assert cs.a.equals(a)
    assert cs.a2.equals(a2)
    assert cs.anonymous.equals(anonymous)
    assert cs.a01.equals(a01)
    assert cs.ano01.equals(ano01)
    assert cs.d == d
    assert cs.e.equals(e)
    assert cs.g.equals(g)


@needs_pytables
def test_init_checkedsession_hdf():
    cs = TestCheckedSession(inputpath('test_session.h5'))
    assert set(cs.keys()) == {'b', 'b024', 'a', 'a2', 'anonymous', 'a01', 'ano01', 'c', 'd', 'e', 'g', 'f', 'h'}


def test_getitem_cs(checkedsession):
    test_getitem(checkedsession)


def test_setitem_cs(checkedsession):
    cs = checkedsession

    # only change values of an array -> OK
    cs['h'] = zeros_like(h)

    # trying to add an undeclared variable -> prints a warning message
    with must_warn(UserWarning, msg=f"'i' is not declared in '{cs.__class__.__name__}'"):
        cs['i'] = ndtest((3, 3))

    # trying to set a variable with an object of different type -> should fail
    # a) type given explicitly
    # -> Axis
    with must_raise(TypeError, msg="instance of Axis expected"):
        cs['a'] = 0
    # -> CheckedArray
    with must_raise(TypeError, msg="Expected object of type 'Array' or a scalar for the variable 'h' but got "
                                   "object of type 'ndarray'"):
        cs['h'] = h.data
    # b) type deduced from the given default value
    with must_raise(TypeError, msg="instance of Axis expected"):
        cs['b'] = ndtest((3, 3))

    # trying to set a CheckedArray variable using a scalar -> OK
    cs['h'] = 5

    # trying to set a CheckedArray variable using an array with axes in different order -> OK
    cs['h'] = h.transpose()
    assert cs.h.axes.names == h.axes.names

    # broadcasting (missing axis) is allowed
    cs['h'] = ndtest(a3)
    assert_array_nan_equal(cs['h']['b0'], cs['h']['b1'])

    # trying to set a CheckedArray variable using an array with wrong axes -> should fail
    # a) extra axis
    with must_raise(ValueError, msg="Array 'h' was declared with axes {a, b} but got array with axes {a, b, c} "
                                    "(unexpected {c} axis)"):
        cs['h'] = ndtest((a3, b2, 'c=c0..c2'))
    # b) incompatible axis
    msg = """\
Incompatible axis for array 'h':
Axis(['a0', 'a1', 'a2', 'a3', 'a4'], 'a')
vs
Axis(['a0', 'a1', 'a2', 'a3'], 'a')"""
    with must_raise(ValueError, msg=msg):
        cs['h'] = h.append('a', 0, 'a4')


def test_getattr_cs(checkedsession):
    test_getattr(checkedsession)


def test_setattr_cs(checkedsession):
    cs = checkedsession

    # only change values of an array -> OK
    cs.h = zeros_like(h)

    # trying to add an undeclared variable -> prints a warning message
    with must_warn(UserWarning, msg=f"'i' is not declared in '{cs.__class__.__name__}'"):
        cs.i = ndtest((3, 3))

    # trying to set a variable with an object of different type -> should fail
    # a) type given explicitly
    # -> Axis
    with must_raise(TypeError, msg="instance of Axis expected"):
        cs.a = 0
    # -> CheckedArray
    with must_raise(TypeError, msg="Expected object of type 'Array' or a scalar for the variable 'h' but got "
                                   "object of type 'ndarray'"):
        cs.h = h.data
    # b) type deduced from the given default value
    with must_raise(TypeError, msg="instance of Axis expected"):
        cs.b = ndtest((3, 3))

    # trying to set a CheckedArray variable using a scalar -> OK
    cs.h = 5

    # trying to set a CheckedArray variable using an array with axes in different order -> OK
    cs.h = h.transpose()
    assert cs.h.axes.names == h.axes.names

    # broadcasting (missing axis) is allowed
    cs.h = ndtest(a3)
    assert_array_nan_equal(cs.h['b0'], cs.h['b1'])

    # trying to set a CheckedArray variable using an array with wrong axes -> should fail
    # a) extra axis
    with must_raise(ValueError, msg="Array 'h' was declared with axes {a, b} but got array with axes {a, b, c} "
                                    "(unexpected {c} axis)"):
        cs.h = ndtest((a3, b2, 'c=c0..c2'))
    # b) incompatible axis
    msg = """\
Incompatible axis for array 'h':
Axis(['a0', 'a1', 'a2', 'a3', 'a4'], 'a')
vs
Axis(['a0', 'a1', 'a2', 'a3'], 'a')"""
    with must_raise(ValueError, msg=msg):
        cs.h = h.append('a', 0, 'a4')


def test_add_cs(checkedsession):
    cs = checkedsession
    test_add(cs)

    u = Axis('u=u0..u2')
    with must_warn(UserWarning, msg=f"'u' is not declared in '{cs.__class__.__name__}'", check_file=False):
        cs.add(u)


def test_iter_cs(checkedsession):
    # As of v1.0 of pydantic all fields with annotations (whether annotation-only or with a default value)
    # will precede all fields without an annotation. Within their respective groups, fields remain in the
    # order they were defined.
    # See https://pydantic-docs.helpmanual.io/usage/models/#field-ordering
    expected = [a, a2, a01, c, e, g, f, h, b, b024, anonymous, ano01, d]
    assert_seq_equal(checkedsession, expected)


def test_filter_cs(checkedsession):
    # see comment in test_iter_cs() about fields ordering
    cs = checkedsession
    cs.ax = 'ax'
    assert_seq_equal(cs.filter(), [a, a2, a01, c, e, g, f, h, b, b024, anonymous, ano01, d, 'ax'])
    assert_seq_equal(cs.filter('a*'), [a, a2, a01, anonymous, ano01, 'ax'])
    assert list(cs.filter('a*', dict)) == []
    assert list(cs.filter('a*', str)) == ['ax']
    assert list(cs.filter('a*', Axis)) == [a, a2, anonymous]
    assert list(cs.filter(kind=Axis)) == [a, a2, b, anonymous]
    assert list(cs.filter('a01', Group)) == [a01]
    assert list(cs.filter(kind=Group)) == [a01, b024, ano01]
    assert_seq_equal(cs.filter(kind=Array), [e, g, f, h])
    assert list(cs.filter(kind=dict)) == [{}]
    assert list(cs.filter(kind=(Axis, Group))) == [a, a2, a01, b, b024, anonymous, ano01]


def test_names_cs(checkedsession):
    assert checkedsession.names == ['a', 'a01', 'a2', 'ano01', 'anonymous', 'b', 'b024',
                                    'c', 'd', 'e', 'f', 'g', 'h']


def _test_io_cs(tmpdir, meta, engine, ext):
    filename = f"test_{engine}.{ext}" if 'csv' not in engine else f"test_{engine}{ext}"
    fpath = tmp_path(tmpdir, filename)

    is_excel_or_csv = 'excel' in engine or 'csv' in engine

    # Save and load
    # -------------

    # a) - all typed variables have a defined value
    #    - no extra variables are added
    csession = TestCheckedSession(a=a, a2=a2, a01=a01, d=d, e=e, g=g, f=f, h=h, meta=meta)
    csession.save(fpath, engine=engine)
    cs = TestCheckedSession()
    cs.load(fpath, engine=engine)
    # --- keys ---
    assert list(cs.keys()) == list(csession.keys())
    # --- variables with default values ---
    assert cs.b.equals(b)
    assert cs.b024.equals(b024)
    assert cs.anonymous.equals(anonymous)
    assert cs.ano01.equals(ano01)
    assert cs.d == d
    # --- typed variables ---
    # Array is support by all formats
    assert cs.e.equals(e)
    assert cs.g.equals(g)
    assert cs.f.equals(f)
    assert cs.h.equals(h)
    # Axis and Group are not supported by the Excel and CSV formats
    if is_excel_or_csv:
        assert isinstance(cs.a, NotLoaded)
        assert isinstance(cs.a2, NotLoaded)
        assert isinstance(cs.a01, NotLoaded)
    else:
        assert cs.a.equals(a)
        assert cs.a2.equals(a2)
        assert cs.a01.equals(a01)
    # --- dtype of Axis variables ---
    if not is_excel_or_csv:
        for key in cs.filter(kind=Axis).keys():
            assert cs[key].dtype == csession[key].dtype
    # --- metadata ---
    if engine != 'pandas_excel':
        assert cs.meta == meta

    # b) - not all typed variables have a defined value
    #    - no extra variables are added
    csession = TestCheckedSession(a=a, d=d, e=e, h=h, meta=meta)
    if 'csv' in engine:
        import shutil
        shutil.rmtree(fpath)
    csession.save(fpath, engine=engine)
    cs = TestCheckedSession()
    cs.load(fpath, engine=engine)
    # --- keys ---
    assert list(cs.keys()) == list(csession.keys())
    # --- variables with default values ---
    assert cs.b.equals(b)
    assert cs.b024.equals(b024)
    assert cs.anonymous.equals(anonymous)
    assert cs.ano01.equals(ano01)
    assert cs.d == d
    # --- typed variables ---
    # Array is support by all formats
    assert cs.e.equals(e)
    assert isinstance(cs.g, NotLoaded)
    assert isinstance(cs.f, NotLoaded)
    assert cs.h.equals(h)
    # Axis and Group are not supported by the Excel and CSV formats
    if is_excel_or_csv:
        assert isinstance(cs.a, NotLoaded)
        assert isinstance(cs.a2, NotLoaded)
        assert isinstance(cs.a01, NotLoaded)
    else:
        assert cs.a.equals(a)
        assert isinstance(cs.a2, NotLoaded)
        assert isinstance(cs.a01, NotLoaded)

    # c) - all typed variables have a defined value
    #    - extra variables are added
    i = ndtest(6)
    j = ndtest((3, 3))
    k = ndtest((2, 2))
    csession = TestCheckedSession(a=a, a2=a2, a01=a01, d=d, e=e, g=g, f=f, h=h, k=k, j=j, i=i, meta=meta)
    csession.save(fpath, engine=engine)
    cs = TestCheckedSession()
    cs.load(fpath, engine=engine)
    # --- names ---
    # we do not use keys() since order of undeclared variables
    # may not be preserved (at least for the HDF format)
    assert cs.names == csession.names
    # --- extra variable ---
    assert cs.i.equals(i)
    assert cs.j.equals(j)
    assert cs.k.equals(k)

    # Update a Group + an Axis + an array (overwrite=False)
    # -----------------------------------------------------
    csession = TestCheckedSession(a=a, a2=a2, a01=a01, d=d, e=e, g=g, f=f, h=h, meta=meta)
    csession.save(fpath, engine=engine)
    a4 = Axis('a=0..3')
    a4_01 = a3['0,1'] >> 'a01'
    e2 = ndtest((a4, 'b=b0..b2'))
    h2 = full_like(h, fill_value=10)
    TestCheckedSession(a=a4, a01=a4_01, e=e2, h=h2).save(fpath, overwrite=False, engine=engine)
    cs = TestCheckedSession()
    cs.load(fpath, engine=engine)
    # --- variables with default values ---
    assert cs.b.equals(b)
    assert cs.b024.equals(b024)
    assert cs.anonymous.equals(anonymous)
    assert cs.ano01.equals(ano01)
    # --- typed variables ---
    # Array is support by all formats
    assert cs.e.equals(e2)
    assert cs.h.equals(h2)
    if engine == 'pandas_excel':
        # Session.save() via engine='pandas_excel' always overwrite the output Excel files
        # arrays 'g' and 'f' have been dropped
        assert isinstance(cs.g, NotLoaded)
        assert isinstance(cs.f, NotLoaded)
        # Axis and Group are not supported by the Excel and CSV formats
        assert isinstance(cs.a, NotLoaded)
        assert isinstance(cs.a2, NotLoaded)
        assert isinstance(cs.a01, NotLoaded)
    elif is_excel_or_csv:
        assert cs.g.equals(g)
        assert cs.f.equals(f)
        # Axis and Group are not supported by the Excel and CSV formats
        assert isinstance(cs.a, NotLoaded)
        assert isinstance(cs.a2, NotLoaded)
        assert isinstance(cs.a01, NotLoaded)
    else:
        assert list(cs.keys()) == list(csession.keys())
        assert cs.a.equals(a4)
        assert cs.a2.equals(a2)
        assert cs.a01.equals(a4_01)
        assert cs.g.equals(g)
        assert cs.f.equals(f)
    if engine != 'pandas_excel':
        assert cs.meta == meta

    # Load only some objects
    # ----------------------
    csession = TestCheckedSession(a=a, a2=a2, a01=a01, d=d, e=e, g=g, f=f, h=h, meta=meta)
    csession.save(fpath, engine=engine)
    cs = TestCheckedSession()
    names_to_load = ['e', 'h'] if is_excel_or_csv else ['a', 'a01', 'a2', 'e', 'h']
    cs.load(fpath, names=names_to_load, engine=engine)
    # --- keys ---
    assert list(cs.keys()) == list(csession.keys())
    # --- variables with default values ---
    assert cs.b.equals(b)
    assert cs.b024.equals(b024)
    assert cs.anonymous.equals(anonymous)
    assert cs.ano01.equals(ano01)
    assert cs.d == d
    # --- typed variables ---
    # Array is support by all formats
    assert cs.e.equals(e)
    assert isinstance(cs.g, NotLoaded)
    assert isinstance(cs.f, NotLoaded)
    assert cs.h.equals(h)
    # Axis and Group are not supported by the Excel and CSV formats
    if is_excel_or_csv:
        assert isinstance(cs.a, NotLoaded)
        assert isinstance(cs.a2, NotLoaded)
        assert isinstance(cs.a01, NotLoaded)
    else:
        assert cs.a.equals(a)
        assert cs.a2.equals(a2)
        assert cs.a01.equals(a01)

    return fpath


@needs_pytables
def test_h5_io_cs(tmpdir, meta):
    _test_io_cs(tmpdir, meta, engine='pandas_hdf', ext='h5')


@needs_openpyxl
def test_xlsx_pandas_io_cs(tmpdir, meta):
    _test_io_cs(tmpdir, meta, engine='pandas_excel', ext='xlsx')


@needs_xlwings
def test_xlsx_xlwings_io_cs(tmpdir, meta):
    _test_io_cs(tmpdir, meta, engine='xlwings_excel', ext='xlsx')


def test_csv_io_cs(tmpdir, meta):
    _test_io_cs(tmpdir, meta, engine='pandas_csv', ext='csv')


def test_pickle_io_cs(tmpdir, meta):
    _test_io_cs(tmpdir, meta, engine='pickle', ext='pkl')


def test_pickle_roundtrip_cs(checkedsession, meta):
    cs = checkedsession
    cs.meta = meta
    s = pickle.dumps(cs)
    res = pickle.loads(s)
    assert res.equals(cs)
    assert res.meta == meta


def test_element_equals_cs(checkedsession):
    test_element_equals(checkedsession)


def test_eq_cs(checkedsession):
    test_eq(checkedsession)


def test_ne_cs(checkedsession):
    test_ne(checkedsession)


def test_sub_cs(checkedsession):
    cs = checkedsession
    session_cls = cs.__class__

    # session - session
    other = session_cls(a=a, a2=a2, a01=a01, e=e - 1, g=zeros_like(g), f=zeros_like(f), h=ones_like(h))
    diff = cs - other
    assert isinstance(diff, session_cls)
    # --- non-array variables ---
    assert diff.b is b
    assert diff.b024 is b024
    assert diff.a is a
    assert diff.a2 is a2
    assert diff.anonymous is anonymous
    assert diff.a01 is a01
    assert diff.ano01 is ano01
    assert diff.c is c
    assert diff.d is d
    # --- array variables ---
    assert_array_nan_equal(diff.e, np.full((2, 3), 1, dtype=np.int32))
    assert_array_nan_equal(diff.g, g)
    assert_array_nan_equal(diff.f, f)
    assert_array_nan_equal(diff.h, h - ones_like(h))

    # session - scalar
    diff = cs - 2
    assert isinstance(diff, session_cls)
    # --- non-array variables ---
    assert diff.b is b
    assert diff.b024 is b024
    assert diff.a is a
    assert diff.a2 is a2
    assert diff.anonymous is anonymous
    assert diff.a01 is a01
    assert diff.ano01 is ano01
    assert diff.c is c
    assert diff.d is d
    # --- non constant arrays ---
    assert_array_nan_equal(diff.e, e - 2)
    assert_array_nan_equal(diff.g, g - 2)
    assert_array_nan_equal(diff.f, f - 2)
    assert_array_nan_equal(diff.h, h - 2)

    # session - dict(Array and scalar)
    other = {'e': ones_like(e), 'h': 1}
    diff = cs - other
    assert isinstance(diff, session_cls)
    # --- non-array variables ---
    assert diff.b is b
    assert diff.b024 is b024
    assert diff.a is a
    assert diff.a2 is a2
    assert diff.anonymous is anonymous
    assert diff.a01 is a01
    assert diff.ano01 is ano01
    assert diff.c is c
    assert diff.d is d
    # --- non constant arrays ---
    assert_array_nan_equal(diff.e, e - ones_like(e))
    assert isnan(diff.g).all()
    assert isnan(diff.f).all()
    assert_array_nan_equal(diff.h, h - 1)

    # session - array
    axes = cs.h.axes
    cs.e = ndtest(axes)
    cs.g = ones_like(cs.h)
    diff = cs - ones(axes)
    assert isinstance(diff, session_cls)
    # --- non-array variables ---
    assert diff.b is b
    assert diff.b024 is b024
    assert diff.a is a
    assert diff.a2 is a2
    assert diff.anonymous is anonymous
    assert diff.a01 is a01
    assert diff.ano01 is ano01
    assert diff.c is c
    assert diff.d is d
    # --- non constant arrays ---
    assert_array_nan_equal(diff.e, cs.e - ones(axes))
    assert_array_nan_equal(diff.g, cs.g - ones(axes))
    assert isnan(diff.f).all()
    assert_array_nan_equal(diff.h, cs.h - ones(axes))


def test_rsub_cs(checkedsession):
    cs = checkedsession
    session_cls = cs.__class__

    # scalar - session
    diff = 2 - cs
    assert isinstance(diff, session_cls)
    # --- non-array variables ---
    assert diff.b is b
    assert diff.b024 is b024
    assert diff.a is a
    assert diff.a2 is a2
    assert diff.anonymous is anonymous
    assert diff.a01 is a01
    assert diff.ano01 is ano01
    assert diff.c is c
    assert diff.d is d
    # --- non constant arrays ---
    assert_array_nan_equal(diff.e, 2 - e)
    assert_array_nan_equal(diff.g, 2 - g)
    assert_array_nan_equal(diff.f, 2 - f)
    assert_array_nan_equal(diff.h, 2 - h)

    # dict(Array and scalar) - session
    other = {'e': ones_like(e), 'h': 1}
    diff = other - cs
    assert isinstance(diff, session_cls)
    # --- non-array variables ---
    assert diff.b is b
    assert diff.b024 is b024
    assert diff.a is a
    assert diff.a2 is a2
    assert diff.anonymous is anonymous
    assert diff.a01 is a01
    assert diff.ano01 is ano01
    assert diff.c is c
    assert diff.d is d
    # --- non constant arrays ---
    assert_array_nan_equal(diff.e, ones_like(e) - e)
    assert isnan(diff.g).all()
    assert isnan(diff.f).all()
    assert_array_nan_equal(diff.h, 1 - h)


def test_neg_cs(checkedsession):
    cs = checkedsession
    neg_cs = -cs
    # --- non-array variables ---
    assert isnan(neg_cs.b)
    assert isnan(neg_cs.b024)
    assert isnan(neg_cs.a)
    assert isnan(neg_cs.a2)
    assert isnan(neg_cs.anonymous)
    assert isnan(neg_cs.a01)
    assert isnan(neg_cs.ano01)
    assert isnan(neg_cs.c)
    assert isnan(neg_cs.d)
    # --- non constant arrays ---
    assert_array_nan_equal(neg_cs.e, -e)
    assert_array_nan_equal(neg_cs.g, -g)
    assert_array_nan_equal(neg_cs.f, -f)
    assert_array_nan_equal(neg_cs.h, -h)


if __name__ == "__main__":
    pytest.main()



import fnmatch
import re
import warnings
from itertools import product, chain

import numpy as np
import pandas as pd

from larray.core.abstractbases import ABCAxis, ABCAxisReference, ABCArray
from larray.util.oset import OrderedSet
from larray.util.misc import (unique_list, find_closing_chr, _parse_bound, _seq_summary, _isintstring, renamed_to,
                              LHDFStore)


def _slice_to_str(key, repr_func=str):
    r"""
    Converts a slice to a string

    Examples
    --------
    >>> _slice_to_str(slice(None))
    ':'
    >>> _slice_to_str(slice(24))
    ':24'
    >>> _slice_to_str(slice(25, None))
    '25:'
    >>> _slice_to_str(slice(5, 10))
    '5:10'
    >>> _slice_to_str(slice(None, 5, 2))
    ':5:2'
    """
    start = repr_func(key.start) if key.start is not None else ''
    stop = repr_func(key.stop) if key.stop is not None else ''
    step = (":" + repr_func(key.step)) if key.step is not None else ''
    return f'{start}:{stop}{step}'


def irange(start, stop, step=None):
    r"""Create a range, with inclusive stop bound and automatic sign for step.

    Parameters
    ----------
    start : int
        Start bound
    stop : int
        Inclusive stop bound
    step : int, optional
        Distance between two generated numbers. If provided this *must* be a positive integer.

    Returns
    -------
    range

    Examples
    --------
    >>> list(irange(1, 3))
    [1, 2, 3]
    >>> list(irange(2, 0))
    [2, 1, 0]
    >>> list(irange(1, 6, 2))
    [1, 3, 5]
    >>> list(irange(6, 1, 2))
    [6, 4, 2]
    >>> list(irange(-1, 1))
    [-1, 0, 1]
    """
    if step is None:
        step = 1
    elif step <= 0:
        raise ValueError("step must be a positive integer or None")
    step = step if start <= stop else -step
    stop = stop + 1 if start <= stop else stop - 1
    return range(start, stop, step)


_range_bound_pattern = re.compile(r'([0-9]+|[a-zA-Z]+)')


def generalized_range(start, stop, step=1):
    r"""Create a range, with inclusive stop bound and automatic sign for step. Bounds can be strings.

    Parameters
    ----------
    start : int or str
        Start bound
    stop : int or str
        Inclusive stop bound
    step : int, optional
        Distance between two generated numbers. If provided this *must* be a positive integer.

    Returns
    -------
    range

    Examples
    --------
    works with both number and letter bounds

    >>> list(generalized_range(-1, 2))
    [-1, 0, 1, 2]
    >>> generalized_range('a', 'c')
    ['a', 'b', 'c']

    can generate in reverse

    >>> list(generalized_range(2, 0))
    [2, 1, 0]
    >>> generalized_range('c', 'a')
    ['c', 'b', 'a']

    can combine letters and numbers

    >>> generalized_range('a0', 'c1')
    ['a0', 'a1', 'b0', 'b1', 'c0', 'c1']

    any special character is left intact

    >>> generalized_range('a_0', 'c_1')
    ['a_0', 'a_1', 'b_0', 'b_1', 'c_0', 'c_1']

    consecutive digits are treated like numbers

    >>> generalized_range('A8', 'A10')
    ['A8', 'A9', 'A10']

    one may use zero padding on numbers

    >>> generalized_range('A08', 'A10')
    ['A08', 'A09', 'A10']

    consecutive letters create all combinations

    >>> generalized_range('AA', 'CC')
    ['AA', 'AB', 'AC', 'BA', 'BB', 'BC', 'CA', 'CB', 'CC']

    one cannot go from a integer to a letter and vice versa

    >>> generalized_range('1', 'F')
    Traceback (most recent call last):
    ...
    ValueError: expected an integer for the stop bound (because the start bound is an integer) but got 'F' instead

    when using special characters, they must be the same on both sides

    >>> generalized_range('a|1', 'a/2')
    Traceback (most recent call last):
    ...
    ValueError: Special characters must be the same for start and stop
    """
    if isinstance(start, str):
        assert isinstance(stop, str)
        start_parts = _range_bound_pattern.split(start)
        stop_parts = _range_bound_pattern.split(stop)
        assert len(start_parts) == len(stop_parts)
        ranges = []
        for start_part, stop_part in zip(start_parts, stop_parts):
            # we only handle non-negative int-like strings on purpose. Int-only bounds should already be converted to
            # real integers by now, and mixing negative int-like strings and letters yields some strange results.
            if start_part.isdigit():
                if not stop_part.isdigit():
                    raise ValueError(f"expected an integer for the stop bound (because the start bound is an integer) "
                                     f"but got '{stop_part}' instead")
                rng = irange(int(start_part), int(stop_part))
                start_pad = len(start_part) if start_part.startswith('0') else None
                stop_pad = len(stop_part) if stop_part.startswith('0') else None
                if start_pad is not None and stop_pad is not None and start_pad != stop_pad:
                    raise ValueError(f"Inconsistent zero padding for start and stop ({start_pad} vs {stop_pad}) of the "
                                     f"numerical part. Must be either the same on both sides or no padding on either "
                                     f"side")
                elif start_pad is None and stop_pad is None:
                    r = [str(num) for num in rng]
                else:
                    pad = start_pad if stop_pad is None else stop_pad
                    r = ['%0*d' % (pad, num) for num in rng]
            elif start_part.isalpha():
                assert stop_part.isalpha()
                int_start = [ord(c) for c in start_part]
                int_stop = [ord(c) for c in stop_part]
                sranges = [[chr(c) for c in irange(r_start, r_stop) if chr(c).isalnum()]
                           for r_start, r_stop in zip(int_start, int_stop)]
                r = [''.join(p) for p in product(*sranges)]
            else:
                # special characters
                if start_part != stop_part:
                    raise ValueError("Special characters must be the same for start and stop")
                r = [start_part]
            ranges.append(r)
        res = [''.join(p) for p in product(*ranges)]
        return res if step == 1 else res[::step]
    else:
        return irange(start, stop, step)


_range_str_pattern = re.compile(r'(?P<start>[^\s.]+)?\s*\.\.\s*(?P<stop>[^\s.]+)?(\s+step\s+(?P<step>\d+))?')


def _range_str_to_range(s, stack_depth=1):
    r"""
    Converts a range string to a range (of values).
    The end point is included.

    Parameters
    ----------
    s : str
        String representing a range of values

    Returns
    -------
    range
        range of int or list of str.

    Examples
    --------
    >>> list(_range_str_to_range('-1..2'))
    [-1, 0, 1, 2]
    >>> _range_str_to_range('a..c')
    ['a', 'b', 'c']
    >>> list(_range_str_to_range('2..6 step 2'))
    [2, 4, 6]

    0 padding
    >>> list(_range_str_to_range('01..12'))
    ['01', '02', '03', '04', '05', '06', '07', '08', '09', '10', '11', '12']

    any special character except . and spaces should work
    >>> _range_str_to_range('a|+*@-b .. a|+*@-d')
    ['a|+*@-b', 'a|+*@-c', 'a|+*@-d']
    """
    s = s.strip()
    m = _range_str_pattern.match(s)

    groups = m.groupdict()
    start, stop, step = groups['start'], groups['stop'], groups['step']
    start = _parse_bound(start, stack_depth + 1) if start is not None else 0
    if stop is None:
        raise ValueError(f"no stop bound provided in range: {s!r}")
    stop = _parse_bound(stop, stack_depth + 1)
    if isinstance(start, str) or isinstance(stop, str):
        start, stop = str(start), str(stop)
    # TODO: use parse_bound
    step = int(step) if step is not None else 1
    return generalized_range(start, stop, step)


def _normalize_idx(idx, length):
    if idx < - length:
        return - length - 1
    elif -length <= idx < 0:
        return idx + length
    elif 0 <= idx < length:
        return idx
    elif idx > length:
        return length


def _idx_seq_to_slice(seq, length):
    r"""
    Transform a sequence of indices into a slice if possible and normalize it.

    If a constant step between indices of the sequence can be inferred (the sequence is something like
    [start, start+step, start+2*step, ...]), a normalized (with has few negative indices as possible) slice is
    returned, otherwise the sequence is returned unmodified.

    Parameters
    ----------
    seq : sequence-like of integers
        Sequence (list, tuple or ndarray) of indices.
    length : int
        Length of the sequence the indices refer to.

    Returns
    -------
    slice or sequence-like
        return the input sequence if a slice cannot be defined.

    Examples
    --------
    >>> _idx_seq_to_slice([3, 4, 5], 10)
    slice(3, 6, None)
    >>> _idx_seq_to_slice([3, 4, 6], 10)
    [3, 4, 6]
    >>> _idx_seq_to_slice([3, 5, 7], 10)
    slice(3, 8, 2)
    >>> _idx_seq_to_slice([-3, -2], 10)
    slice(7, 9, None)
    >>> _idx_seq_to_slice([-1, -2], 10)
    slice(9, 7, -1)
    >>> _idx_seq_to_slice([2, 1], 10)
    slice(2, 0, -1)
    >>> _idx_seq_to_slice([1, 0], 10)
    slice(1, -11, -1)
    >>> _idx_seq_to_slice([1], 10)
    slice(1, 2, None)
    >>> _idx_seq_to_slice([], 10)
    []
    >>> _idx_seq_to_slice([3, 1], 10)
    slice(3, 0, -2)
    """
    if len(seq) < 1:
        return seq
    first_idx = _normalize_idx(seq[0], length)
    if len(seq) == 1:
        return slice(first_idx, first_idx + 1)
    second_idx = _normalize_idx(seq[1], length)
    step = second_idx - first_idx
    prev_idx = second_idx
    for raw_idx in seq[2:]:
        idx = _normalize_idx(raw_idx, length)
        if idx != prev_idx + step:
            return seq
        prev_idx = idx
    last_idx = prev_idx
    step_sign = 1 if step > 0 else -1
    # we stop just after last_idx instead of using a full "step" because for negative steps, it could get us a negative
    # stop which isn't what we want
    stop = last_idx + step_sign
    if last_idx == 0 and step < 0:
        stop -= length
    if step == 1:
        step = None
    return slice(first_idx, stop, step)


def _is_object_array(array):
    return isinstance(array, np.ndarray) and array.dtype.type == np.object_


def _can_have_groups(seq):
    return _is_object_array(seq) or isinstance(seq, (tuple, list))


def _contain_group_ticks(ticks):
    return _can_have_groups(ticks) and any(isinstance(tick, Group) for tick in ticks)


def _seq_group_to_name(seq):
    if _can_have_groups(seq):
        return [v.name if isinstance(v, Group) else v for v in seq]
    else:
        return seq


def _to_tick(v):
    r"""
    Converts any value to a tick (ie makes it hashable, and acceptable as an ndarray element)

    scalar -> not modified
    slice -> 'start:stop'
    list|tuple -> 'v1,v2,v3'
    Group with name -> v.name
    Group without name -> _to_tick(v.key)
    other -> str(v)

    Parameters
    ----------
    v : any
        value to be converted.

    Returns
    -------
    any scalar
        scalar representing the tick
    """
    # the fact that an "aggregated tick" is passed as a LGroup or as a string should be as irrelevant as possible.
    # The thing is that we cannot (currently) use the more elegant _to_tick(e.key) that means the LGroup is not
    # available in Axis.__init__ after to_ticks, and we need it to update the mapping if it was named. Effectively,
    # this creates two entries in the mapping for a single tick. Besides, I like having the LGroup as the tick, as it
    # provides extra info as to where it comes from.
    if np.isscalar(v):
        return v
    elif isinstance(v, Group):
        return v.name if v.name is not None else _to_tick(v.to_label())
    elif isinstance(v, slice):
        return _slice_to_str(v)
    elif isinstance(v, (tuple, list)):
        if len(v) == 1:
            return str(v) + ','
        else:
            # TODO: it would be nicer/saner to use n=1, sep='' but this currently breaks at lot of tests
            return _seq_summary(v, n=1000, repr_func=str, sep=',')
    else:
        return str(v)


def _to_ticks(s, parse_single_int=False):
    r"""
    Makes a (list of) value(s) usable as the collection of labels for an Axis (ie hashable).

    Strip strings, split them on ',' and translate "range strings" to list of values **including the end point** !

    Parameters
    ----------
    s : iterable
        List of values usable as the collection of labels for an Axis.

    Returns
    -------
    Numpy 1D array containing labels

    Examples
    --------
    >>> list(_to_ticks('M , F'))                # doctest: +NORMALIZE_WHITESPACE
    ['M', 'F']
    >>> list(_to_ticks('A,C..E,F..G,Z'))        # doctest: +NORMALIZE_WHITESPACE
    ['A', 'C', 'D', 'E', 'F', 'G', 'Z']
    >>> list(_to_ticks('U'))                    # doctest: +NORMALIZE_WHITESPACE
    ['U']
    >>> list(_to_ticks('..3'))                  # doctest: +NORMALIZE_WHITESPACE
    [0, 1, 2, 3]
    >>> list(_to_ticks('01..12'))               # doctest: +NORMALIZE_WHITESPACE
    ['01', '02', '03', '04', '05', '06', '07', '08', '09', '10', '11', '12']
    >>> list(_to_ticks('01,02,03,10,11,12'))    # doctest: +NORMALIZE_WHITESPACE
    ['01', '02', '03', '10', '11', '12']
    """
    if isinstance(s, ABCAxis):
        return s.labels
    if isinstance(s, Group):
        # a single LGroup used for all ticks of an Axis
        return _to_ticks(s.eval())
    elif isinstance(s, np.ndarray):
        # we assume it has already been translated
        # XXX: Is it a safe assumption?
        return s

    if isinstance(s, pd.Index):
        ticks = s.values
    elif isinstance(s, (list, tuple)):
        ticks = [_to_tick(e) for e in s]
    elif isinstance(s, range):
        ticks = s
    elif isinstance(s, str):
        seq = _seq_str_to_seq(s, parse_single_int=parse_single_int)
        if isinstance(seq, slice):
            raise ValueError("using : to define axes is deprecated, please use .. instead")
        ticks = [seq] if isinstance(seq, (str, int)) else seq
    elif hasattr(s, '__array__'):
        ticks = s.__array__()
    else:
        try:
            ticks = list(s)
        except TypeError:
            raise TypeError(f"ticks must be iterable ({type(s)} is not)")
    return np.asarray(ticks)


_axis_name_pattern = re.compile(r'\s*(([A-Za-z0-9]\w*)(\.i)?\s*\[)?(.*)')


def _seq_str_to_seq(s, stack_depth=1, parse_single_int=False):
    r"""
    Converts a sequence string to its sequence (or scalar)

    Parameters
    ----------
    s : str
        string to parse

    Returns
    -------
    scalar, slice, range or list
    """
    numcolons = s.count(':')
    if numcolons:
        assert numcolons <= 2
        # bounds can be of len 2 or 3 (if step is provided)
        # stack_depth + 2 because the list comp has its own stack
        bounds = [_parse_bound(b, stack_depth + 2) for b in s.split(':')]
        return slice(*bounds)
    elif ',' in s and '..' in s:
        # strip extremity commas to avoid empty string sequence elements
        s = s.strip(',')

        def to_seq(b, stack_depth=1):
            if '..' in b:
                return _range_str_to_range(b, stack_depth + 1)
            else:
                parsed = _parse_bound(b, stack_depth + 1)
                return (parsed,)

        # stack_depth + 2 because the list comp has its own stack
        return list(chain(*[to_seq(b, stack_depth + 2) for b in s.split(',')]))
    elif ',' in s:
        # strip extremity commas to avoid empty string sequence elements
        s = s.strip(',')
        return [_parse_bound(b, stack_depth + 2) for b in s.split(',')]
    elif '..' in s:
        return _range_str_to_range(s, stack_depth + 1)
    else:
        return _parse_bound(s, stack_depth + 1, parse_int=parse_single_int)


def _to_key(v, stack_depth=1, parse_single_int=False):
    r"""
    Converts a value to a key usable for indexing (slice object, list of values,...).
    Strings are split on ',' and stripped. Colons (:) are interpreted as slices.

    Parameters
    ----------
    v : int or str or tuple or list or slice or Array or Group
        value to convert into a key usable for indexing

    Returns
    -------
    key
        a key represents any object that can be used for indexing

    Examples
    --------
    >>> _to_key('a:c')
    slice('a', 'c', None)
    >>> _to_key('a, b,c ,')
    ['a', 'b', 'c']
    >>> _to_key('a..c')
    ['a', 'b', 'c']
    >>> _to_key('a,c..e,g..h,z')
    ['a', 'c', 'd', 'e', 'g', 'h', 'z']
    >>> _to_key('a,')
    ['a']
    >>> _to_key(' a ')
    'a'
    >>> _to_key(10)
    10
    >>> _to_key('10')
    '10'
    >>> _to_key('10:20')
    slice(10, 20, None)
    >>> _to_key(slice('10', '20'))
    slice('10', '20', None)
    >>> _to_key('year.i[-1]')
    year.i[-1]
    >>> _to_key('age[10:19]>>teens')
    age[10:19] >> 'teens'
    >>> _to_key('a,b,c >> abc')
    LGroup(['a', 'b', 'c']) >> 'abc'
    >>> _to_key('a:c >> abc')
    LGroup(slice('a', 'c', None)) >> 'abc'
    >>> _to_key('0[a0]')
    0['a0']
    >>> _to_key('0.i[a0]')
    0.i['a0']
    >>> ext = [1, 2, 3]
    >>> _to_key('{ext} >> ext')
    LGroup([1, 2, 3]) >> 'ext'
    >>> answer = 42
    >>> _to_key('{answer}')
    42
    >>> _to_key('{answer} >> answer')
    LGroup(42) >> 'answer'
    >>> _to_key('10:{answer} >> answer')
    LGroup(slice(10, 42, None)) >> 'answer'
    >>> _to_key('4,{answer},2 >> answer')
    LGroup([4, 42, 2]) >> 'answer'
    >>> list(_to_key('40..{answer}'))
    [40, 41, 42]
    >>> _to_key('4,40..{answer},2')
    [4, 40, 41, 42, 2]
    >>> _to_key('4,40..{answer},2 >> answer')
    LGroup([4, 40, 41, 42, 2]) >> 'answer'
    """
    if isinstance(v, tuple):
        return list(v)
    elif isinstance(v, str):
        # axis name
        m = _axis_name_pattern.match(v)
        _, axis, positional, key = m.groups()
        # group name. using rfind in the unlikely case there is another >>
        name_pos = key.rfind('>>')
        name = None
        if name_pos != -1:
            key, name = key[:name_pos].strip(), key[name_pos + 2:].strip()
        if axis is not None:
            axis = axis.strip()
            if _isintstring(axis):
                axis = int(axis)
            axis_bracket_open = m.end(1) - 1
            # check that the string parentheses are correctly balanced
            _ = find_closing_chr(v, axis_bracket_open)
            # strip closing bracket (it should be at the end because we took care of the name earlier)
            assert key[-1] == ']'
            key = key[:-1]
        if name is not None or axis is not None:
            cls = IGroup if positional else LGroup
            key = _to_key(key, stack_depth + 1, parse_single_int=positional)
            return cls(key, name=name, axis=axis)
        else:
            return _seq_str_to_seq(v, stack_depth + 1, parse_single_int=parse_single_int)
    elif v is Ellipsis or np.isscalar(v) or isinstance(v, (Group, slice, list, np.ndarray, ABCArray, OrderedSet)):
        return v
    else:
        raise TypeError(f"{v} has an invalid type ({type(v).__name__}) for a key")


def _to_keys(value, stack_depth=1):
    r"""
    Converts a (collection of) group(s) to a structure usable for indexing.

    'label' or ['l1', 'l2'] or [['l1', 'l2'], ['l3']]

    Parameters
    ----------
    value : int or str or tuple or list or slice or Array or Group
        (collection of) value(s) to convert into key(s) usable for indexing

    Returns
    -------
    list of keys

    Examples
    --------
    It is only used for .sum(axis=xxx)
    >>> _to_keys('P01,P02')  # <-- one group => collapse dimension
    ['P01', 'P02']
    >>> _to_keys(('P01,P02',))  # <-- do not collapse dimension
    (['P01', 'P02'],)
    >>> _to_keys('P01;P02,P03;:')
    ('P01', ['P02', 'P03'], slice(None, None, None))

    >>> ext = 'P03'
    >>> _to_keys('P01,P02,{ext}')
    ['P01', 'P02', 'P03']
    >>> _to_keys('P01;P02;{ext}')
    ('P01', 'P02', 'P03')

    >>> _to_keys('age[10:19] >> teens ; year.i[-1]')
    (age[10:19] >> 'teens', year.i[-1])

    # >>> to_keys('P01,P02,:') # <-- INVALID !
    # it should have an explicit failure

    # we allow this, even though it is a dubious syntax
    >>> _to_keys(('P01', 'P02', ':'))
    ('P01', 'P02', slice(None, None, None))

    # it is better to use explicit groups
    >>> _to_keys(('P01,', 'P02,', ':'))
    (['P01'], ['P02'], slice(None, None, None))

    # or even the ugly duck...
    >>> _to_keys((('P01',), ('P02',), ':'))
    (['P01'], ['P02'], slice(None, None, None))
    """
    if isinstance(value, str) and ';' in value:
        value = tuple(value.split(';'))

    if isinstance(value, tuple):
        # stack_depth + 2 because the list comp has its own stack
        return tuple([_to_key(group, stack_depth + 2) for group in value])
    else:
        return _to_key(value, stack_depth + 1)


# forbidden characters in sheet names
_sheet_name_pattern = re.compile(r'[\\/?*\[\]:]')


def _translate_sheet_name(sheet_name):
    if isinstance(sheet_name, Group):
        sheet_name = str(_to_tick(sheet_name))
    if isinstance(sheet_name, str):
        sheet_name = _sheet_name_pattern.sub('_', sheet_name)
        if len(sheet_name) > 31:
            raise ValueError("Sheet names cannot exceed 31 characters")
    return sheet_name


# forbidden characters for dataset names in HDF files
_key_hdf_pattern = re.compile(r'[\\/]')


def _translate_group_key_hdf(key):
    if isinstance(key, Group):
        key = _key_hdf_pattern.sub('_', str(_to_tick(key)))
    return key


# TODO: kill this function
def union(*args):
    # TODO: add support for LGroup and lists
    r"""
    Returns the union of several "value strings" as a list.

    Parameters
    ----------
    *args
        (collection of) value(s) to be converted into label(s). Repeated values are taken only once.

    Returns
    -------
    list of labels

    Examples
    --------
    >>> union('a', 'a, b, c, d', ['d', 'e', 'f'], '..2')
    ['a', 'b', 'c', 'd', 'e', 'f', 0, 1, 2]
    """
    if args:
        return unique_list(chain(*(_to_ticks(arg) for arg in args)))
    else:
        return []


class IGroupMaker:
    r"""
    Generates a new instance of IGroup for a given axis and key.

    Attributes
    ----------
    axis : Axis
        an axis.

    Notes
    -----
    This class is used by the method `Axis.i`
    """
    __slots__ = ('axis',)

    def __init__(self, axis):
        assert isinstance(axis, ABCAxis)
        self.axis = axis

    def __getitem__(self, key):
        if isinstance(key, (int, np.integer)) and not isinstance(self.axis, ABCAxisReference) and key >= len(self.axis):
            raise IndexError(f"{key} is out of range for axis of length {len(self.axis)}")
        return IGroup(key, None, self.axis)

    def __len__(self):
        return len(self.axis)


# We need a separate class for LGroup and cannot simply create a new Axis with a subset of values/ticks/labels:
# the subset of ticks/labels of the LGroup need to correspond to its *Axis* indices
class Group:
    """Abstract Group.
    """
    __slots__ = ('key', 'name', 'axis')

    format_string = None

    def __init__(self, key, name=None, axis=None):
        if isinstance(key, tuple):
            key = list(key)
        if isinstance(key, Group):
            key = key.to_label()
        self.key = remove_nested_groups(key)

        # we do NOT assign a name automatically when missing because that makes it impossible to know whether a name
        # was explicitly given or not
        self.name = _to_tick(name) if name is not None else name
        assert axis is None or isinstance(axis, (str, int, ABCAxis)), f"invalid axis '{axis}' ({type(axis).__name__})"

        # we could check the key is valid but this can be slow and could be useless
        # TODO: for performance reasons, we should cache the result. This will need to be invalidated correctly
        # axis.translate(key)

        # we store the Axis object and not its name like we did previously so that groups on anonymous axes are more
        # meaningful and that we can iterate on a slice of an axis (an LGroup). The reason to store the name instead of
        # the object was to make sure that a Group from an axis (or without axis) could be used on another axis with
        # the same name. See test_array.py:test_...
        self.axis = axis

    def __repr__(self):
        key = self.key

        # eval only returns a slice for groups without an Axis object
        if isinstance(key, slice):
            key_repr = _slice_to_str(key, repr_func=repr)
        elif isinstance(key, (tuple, list, np.ndarray, OrderedSet)):
            key_repr = _seq_summary(key, n=1000, repr_func=repr, sep=', ')
        else:
            key_repr = repr(key)

        axis_name = self.axis.name if isinstance(self.axis, ABCAxis) else self.axis
        if axis_name is not None:
            axis_name = f'X.{axis_name}' if isinstance(self.axis, ABCAxisReference) else axis_name
            s = self.format_string.format(axis=axis_name, key=key_repr)
        else:
            if self.axis is not None:
                # anonymous axis
                axis_ref = f', axis={repr(self.axis)}'
            else:
                axis_ref = ''

            if isinstance(key, slice):
                key_repr = repr(key)
            elif isinstance(key, list):
                key_repr = f'[{key_repr}]'

            s = f'{self.__class__.__name__}({key_repr}{axis_ref})'
        return f"{s} >> {repr(self.name)}" if self.name is not None else s

    def __str__(self):
        return str(self.eval())

    # TODO: rename to "to_positional"
    def translate(self, bound=None, stop=False):
        r"""
        Translate key to a position if it is not already

        Parameters
        ----------
        bound : any, optional
        stop : bool, optional

        Returns
        -------
        int-based key (single int, slice of int or tuple/list/array of them)
        """
        raise NotImplementedError()

    def eval(self):
        r"""
        Translate key to labels, if it is not already, expanding slices in the process.

        Returns
        -------
        label-based key (single scalar or tuple/list/array of them)
        """
        raise NotImplementedError()

    def to_label(self):
        r"""
        Translate key to labels, if it is not already

        Returns
        -------
        label-based key (single scalar, slice of scalars or tuple/list/array of them)
        """
        raise NotImplementedError()

    def retarget_to(self, target_axis):
        r"""Retarget group to another axis.

        It will be translated to an LGroup using its former axis, if necessary.

        Parameters
        ----------
        target_axis : Axis
            axis to conform to

        Returns
        -------
        Group with axis, raise ValueError if retargeting is not possible
        """
        assert isinstance(target_axis, ABCAxis)
        if self.axis is target_axis:
            return self
        elif isinstance(self.axis, str) or isinstance(self.axis, ABCAxisReference):
            axis_name = self.axis.name if isinstance(self.axis, ABCAxisReference) else self.axis
            if axis_name != target_axis.name:
                raise ValueError('cannot retarget a Group defined without a real axis object (e.g. using '
                                 'an AxisReference (X.)) to an axis with a different name')
            return self.__class__(self.key, self.name, target_axis)
        elif isinstance(self.axis, int) or self.axis.equals(target_axis):
            # in the case of isinstance(self.axis, int), we can only hope the axis corresponds. This is the
            # case if we come from _translate_axis_key_chunk, but if the users calls this manually, we cannot know.
            # XXX: maybe changing this to retarget_to_axes would be a good idea after all?

            # just change the axis object
            return self.__class__(self.key, self.name, target_axis)
        else:
            # to retarget to another (non-equal) Axis, we need to translate to labels and expand slices
            return LGroup(self.eval(), self.name, target_axis)

    def __len__(self):
        # XXX: we probably want to_label instead of .eval (so that we do not expand slices)
        value = self.eval()
        # for some reason this breaks having LGroup ticks/labels on an axis
        # if isinstance(value, (tuple, list, Array, np.ndarray, str)):
        if hasattr(value, '__len__'):
            return len(value)
        elif isinstance(value, slice):
            start, stop = value.start, value.stop
            # not using stop - start because that does not work for string bounds
            # (and it is different for LGroup & IGroup)
            start_pos = self.translate(start)
            stop_pos = self.translate(stop)
            return stop_pos - start_pos
        else:
            raise TypeError(f'len() of unsized object ({value})')

    def __iter__(self):
        # XXX: use translate/IGroup instead, so that it works even in the presence of duplicate labels
        #      possibly, only if axis is set?
        axis = self.axis
        return iter([LGroup(v, axis=axis) for v in self.eval()])

    def named(self, name):
        r"""Returns group with a different name.

        Parameters
        ----------
        name : str
            new name for group

        Returns
        -------
        Group
        """
        return self.__class__(self.key, name, self.axis)
    __rshift__ = named

    def with_axis(self, axis):
        r"""Returns group with a different axis.

        Parameters
        ----------
        axis : int, str, Axis
            new axis for group

        Returns
        -------
        Group
        """
        return self.__class__(self.key, self.name, axis)

    def by(self, length, step=None, template=None):
        r"""Split group into several groups of specified length.

        Parameters
        ----------
        length : int
            length of new groups
        step : int, optional
            step between groups. Defaults to length.
        template : str, optional
            template describing how group names are generated. It is a string containing specific arguments
            written inside brackets {}. Available arguments are {start} and {end} representing the first and last label
            of each group. By default, template is defined as '{start}:{end}'.

        Notes
        -----
        step can be smaller than length, in which case, this will produce overlapping groups.

        Returns
        -------
        list of Group

        Examples
        --------
        >>> from larray import Axis, X
        >>> age = Axis('age=0..100')
        >>> young_children = age[0:6]
        >>> young_children.by(3)
        (age.i[0:3] >> '0:2', age.i[3:6] >> '3:5', age.i[6:7] >> '6')
        >>> young_children.by(3, step=2)
        (age.i[0:3] >> '0:2', age.i[2:5] >> '2:4', age.i[4:7] >> '4:6', age.i[6:7] >> '6')
        >>> young_children.by(3, template='{start}-{end}')
        (age.i[0:3] >> '0-2', age.i[3:6] >> '3-5', age.i[6:7] >> '6')
        """
        def make_group(start, length, name_template):
            g = self[start:start + length]
            labels = g.eval()
            g.name = name_template.format(start=labels[0], end=labels[-1]) if len(labels) > 1 else str(labels[0])
            return g

        if step is None:
            step = length
        if template is None:
            template = '{start}:{end}'
        return tuple(make_group(start, length, template) for start in range(0, len(self), step))

    # TODO: __getitem__ should work by label and .i[] should work by position. I guess it would be more consistent this
    # way even if the usefulness of subsetting a group with labels is dubious (but it is sometimes practical to treat
    # the group as if it was an axis).
    # >>> vla = geo['...']
    # >>> # first 10 regions of flanders (this could have some use)
    # >>> vla.i[:10]  # => IGroup on geo
    # >>> vla["antwerp", "gent"]  # => LGroup on geo

    # LGroup[] => LGroup
    # IGroup[] => LGroup
    # IGroup.i[] => IGroup
    # LGroup.i[] => IGroup
    def __getitem__(self, key):
        r"""

        Parameters
        ----------
        key : int, slice of int or list of int
            position-based key (even for LGroup)

        Returns
        -------
        Group
        """
        cls = self.__class__
        orig_key = self.key
        # XXX: unsure we should support tuple
        if isinstance(orig_key, (tuple, list)):
            return cls(orig_key[key], None, self.axis)
        elif isinstance(orig_key, slice):
            orig_start, orig_stop, orig_step = orig_key.start, orig_key.stop, orig_key.step
            if orig_step is None:
                orig_step = 1

            orig_start_pos = self.translate(orig_start) if orig_start is not None else 0
            if isinstance(key, slice):
                key_start, key_stop, key_step = key.start, key.stop, key.step
                if key_step is None:
                    key_step = 1

                orig_stop_pos = self.translate(orig_stop, stop=True) if orig_stop is not None else len(self)
                new_start = orig_start_pos + key_start * orig_step
                new_stop = min(orig_start_pos + key_stop * orig_step, orig_stop_pos)
                new_step = orig_step * key_step
                if new_step == 1:
                    new_step = None
                return IGroup(slice(new_start, new_stop, new_step), None, self.axis)
            elif isinstance(key, int):
                return IGroup(orig_start_pos + key * orig_step, None, self.axis)
            elif isinstance(key, (tuple, list)):
                return IGroup([orig_start_pos + k * orig_step for k in key], None, self.axis)
        elif isinstance(orig_key, ABCArray):
            # XXX: why .i ?
            return cls(orig_key.i[key], None, self.axis)
        elif np.isscalar(orig_key):
            # give the opportunity to subset the label/key itself (for example for string keys)
            value = self.eval()
            return value[key]
        else:
            raise TypeError(f"cannot take a subset of {self.key} because it has a '{type(self.key)}' key")

    def _ipython_key_completions_(self):
        return list(self.eval())

    # method factory
    def _binop(opname):
        op_fullname = f'__{opname}__'

        # TODO: implement this in a delayed fashion for axes references
        def opmethod(self, other):
            other_value = other.eval() if isinstance(other, Group) else other
            return getattr(self.eval(), op_fullname)(other_value)

        opmethod.__name__ = op_fullname
        return opmethod

    __matmul__ = _binop('matmul')
    __ror__ = _binop('ror')
    __or__ = _binop('or')
    __rxor__ = _binop('rxor')
    __xor__ = _binop('xor')
    __rand__ = _binop('rand')
    __and__ = _binop('and')
    __rpow__ = _binop('rpow')
    __pow__ = _binop('pow')
    __rdivmod__ = _binop('rdivmod')
    __divmod__ = _binop('divmod')
    __rmod__ = _binop('rmod')
    __mod__ = _binop('mod')
    # div and rdiv are not longer used on Python3+
    __rfloordiv__ = _binop('rfloordiv')
    __floordiv__ = _binop('floordiv')
    __rtruediv__ = _binop('rtruediv')
    __truediv__ = _binop('truediv')
    __rmul__ = _binop('rmul')
    __mul__ = _binop('mul')
    __rsub__ = _binop('rsub')
    __sub__ = _binop('sub')
    __radd__ = _binop('radd')
    __add__ = _binop('add')

    __ge__ = _binop('ge')
    __gt__ = _binop('gt')
    __le__ = _binop('le')
    __lt__ = _binop('lt')

    # having ne and eq use .eval on a slice group creates an ndarray, for which __eq__ does not return a single value,
    # which means, it cannot be in a mapping/Axis, but this is no longer a problem, since we do not create axes with
    # LGroup labels anymore anyway
    __ne__ = _binop('ne')
    __eq__ = _binop('eq')

    def equals(self, other):
        r"""
        Checks if this group is equal to another group.
        Two groups are equal if they have the same group and axis names and correspond to the same labels.

        Parameters
        ----------
        other : Group
            Group to compare with.

        Returns
        -------
        bool
            True if the other group is equal to this group, False otherwise.

        Examples
        --------
        >>> from larray import Axis
        >>> a = Axis('a=a0..a3')
        >>> a02 = a['a0:a2'] >> 'group_a'

        Same group names, axis names and labels

        >>> a02.equals(a02)
        True

        Different group names (one is None)

        >>> a02.equals(a['a0:a2'])
        False

        Different axis name

        >>> other_axis = a.rename('other_name')
        >>> a02.equals(other_axis['a0:a2'] >> 'group_a')
        False

        Different labels

        >>> a02.equals(a['a1:a3'] >> 'group_a')
        False

        Mixing slice and list groups

        >>> a['a0:a2'].equals(a['a0,a1,a2'])
        True

        Mixing LGroup and IGroup

        >>> a['a0:a2'].equals(a.i[0:3])
        True
        """
        if self is other:
            return True
        if self.name != other.name:
            return False
        self_axis_name = self.axis.name if isinstance(self.axis, ABCAxis) else self.axis
        other_axis_name = other.axis.name if isinstance(other.axis, ABCAxis) else other.axis
        if self_axis_name != other_axis_name:
            return False
        res = self == other
        return res if isinstance(res, bool) else all(res)

    def set(self):
        r"""Creates LSet from this group

        Returns
        -------
        LSet
        """
        return LSet(self.eval(), self.name, self.axis)

    def union(self, other):
        r"""Returns (set) union of this label group and other.

        Labels relative order will be kept intact, but only unique labels will be returned. Labels from this group will
        be before labels from other.

        Parameters
        ----------
        other : Group or any sequence of labels
            other labels

        Returns
        -------
        LSet

        Examples
        --------
        >>> from larray import Axis
        >>> a = Axis('a=a0..a2')
        >>> a['a0', 'a1'].union(a['a1', 'a2'])
        a['a0', 'a1', 'a2'].set()
        >>> a['a0', 'a1'].union('a1,a2')
        a['a0', 'a1', 'a2'].set()
        >>> a['a0', 'a1'].union(['a1', 'a2'])
        a['a0', 'a1', 'a2'].set()
        """
        return self.set().union(other)

    def intersection(self, other):
        r"""Returns (set) intersection of this label group and other.

        In other words, this will return labels from this group which are also in other. Labels relative order will be
        kept intact, but only unique labels will be returned.

        Parameters
        ----------
        other : Group or any sequence of labels
            other labels

        Returns
        -------
        LSet

        Examples
        --------
        >>> from larray import Axis
        >>> a = Axis('a=a0..a2')
        >>> a['a0', 'a1'].intersection(a['a1', 'a2'])
        a['a1'].set()
        >>> a['a0', 'a1'].intersection('a1,a2')
        a['a1'].set()
        >>> a['a0', 'a1'].intersection(['a1', 'a2'])
        a['a1'].set()
        """
        return self.set().intersection(other)

    def difference(self, other):
        r"""Returns (set) difference of this label group and other.

        In other words, this will return labels from this group without those in other. Labels relative order will be
        kept intact, but only unique labels will be returned.

        Parameters
        ----------
        other : Group or any sequence of labels
            other labels

        Returns
        -------
        LSet

        Examples
        --------
        >>> from larray import Axis
        >>> a = Axis('a=a0..a2')
        >>> a['a0', 'a1'].difference(a['a1', 'a2'])
        a['a0'].set()
        >>> a['a0', 'a1'].difference('a1,a2')
        a['a0'].set()
        >>> a['a0', 'a1'].difference(['a1', 'a2'])
        a['a0'].set()
        """
        return self.set().difference(other)

    def __contains__(self, item):
        if isinstance(item, Group):
            item = item.eval()
        return item in self.eval()

    def startingwith(self, prefix):
        r"""
        Returns a group with the labels starting with the specified string.

        Parameters
        ----------
        prefix : str or Group
            The prefix to search for.

        Returns
        -------
        LGroup
            Group containing all the labels starting with the given string.

        Examples
        --------
        >>> from larray import Axis
        >>> people = Axis(['Bruce Wayne', 'Arthur Dent', 'Harvey Dent'], 'people')
        >>> group = people.endingwith('Dent')
        >>> group
        people['Arthur Dent', 'Harvey Dent']
        >>> group.startingwith('Art')
        people['Arthur Dent']
        """
        if isinstance(prefix, Group):
            prefix = prefix.eval()
        return LGroup([v for v in self.eval() if v.startswith(prefix)], axis=self.axis)

    def endingwith(self, suffix):
        r"""
        Returns a group with the labels ending with the specified string.

        Parameters
        ----------
        suffix : str or Group
            The suffix to search for.

        Returns
        -------
        LGroup
            Group containing all the labels ending with the given string.

        Examples
        --------
        >>> from larray import Axis
        >>> people = Axis(['Bruce Wayne', 'Bruce Willis', 'Arthur Dent'], 'people')
        >>> group = people.startingwith('Bru')
        >>> group
        people['Bruce Wayne', 'Bruce Willis']
        >>> people.endingwith('yne')
        people['Bruce Wayne']
        """
        if isinstance(suffix, Group):
            suffix = suffix.eval()
        return LGroup([v for v in self.eval() if v.endswith(suffix)], axis=self.axis)

    def matching(self, deprecated=None, pattern=None, regex=None):
        r"""
        Returns a group with all the labels matching the specified pattern or regular expression.

        Parameters
        ----------
        pattern : str or Group
            Pattern to match.

            - `?`     matches any single character
            - `*`     matches any number of characters
            - [seq]   matches any character in seq
            - [!seq]  matches any character not in seq

            To match any of the special characters above, wrap the character in brackets. For example, `[?]` matches
            the character `?`.
        regex : str or Group
            Regular expression pattern to match. Regular expressions are more powerful than what the simple patterns
            supported by the `pattern` argument but are also more complex to write.
            See `Regular Expression <https://docs.python.org/3/library/re.html>`_ for more details about how to build
            a regular expression pattern.

        Returns
        -------
        LGroup
            Group containing all the labels matching the pattern.

        Examples
        --------
        >>> from larray import Axis
        >>> people = Axis(['Bruce Wayne', 'Bruce Willis', 'Arthur Dent'], 'people')

        Let us create a group with all names starting with B

        >>> group = people.startingwith('B')
        >>> group
        people['Bruce Wayne', 'Bruce Willis']

        Within that group, all labels containing any characters then W then any characters then s are given by

        >>> group.matching(pattern='*W*s')
        people['Bruce Willis']

        Regular expressions are more powerful but usually harder to write and less readable. For example,
        here are the labels not containing the letter "i".

        >>> group.matching(regex='^[^i]*$')
        people['Bruce Wayne']
        """
        if deprecated is not None:
            assert pattern is None and regex is None
            regex = deprecated
            warnings.warn("Group.matching() first argument will change to `pattern` in a later release. "
                          "If your pattern is a regular expression, use Group.matching(regex='yourpattern')."
                          "If your pattern is a 'simple pattern', use Group.matching(pattern='yourpattern').",
                          FutureWarning, stacklevel=2)
        if pattern is not None and regex is not None:
            raise ValueError("Cannot use both `pattern` and `regex` arguments at the same time in Group.matching()")
        if pattern is None and regex is None:
            raise ValueError("Must provide either `pattern` or `regex` argument in Group.matching()")
        if isinstance(regex, Group):
            regex = regex.eval()
        if pattern is not None:
            if isinstance(pattern, Group):
                pattern = pattern.eval()
            regex = fnmatch.translate(pattern)
        match = re.compile(regex).match
        return LGroup([v for v in self.eval() if match(v)], axis=self.axis)

    def containing(self, substring):
        r"""
        Returns a group with all the labels containing the specified substring.

        Parameters
        ----------
        substring : str or Group
            The substring to search for.

        Returns
        -------
        LGroup
            Group containing all the labels containing the substring.

        Examples
        --------
        >>> from larray import Axis
        >>> people = Axis(['Bruce Wayne', 'Bruce Willis', 'Arthur Dent'], 'people')
        >>> group = people.startingwith('Bru')
        >>> group
        people['Bruce Wayne', 'Bruce Willis']
        >>> group.containing('Will')
        people['Bruce Willis']
        """
        if isinstance(substring, Group):
            substring = substring.eval()
        return LGroup([v for v in self.eval() if substring in v], axis=self.axis)

    def to_hdf(self, filepath, key=None, axis_key=None):
        r"""
        Writes group to a HDF file.

        A HDF file can contain multiple groups.
        The 'key' parameter is a unique identifier for the group.
        The 'axis_key' parameter is the unique identifier for the associated axis.
        The associated axis will be saved if not already present in the HDF file.

        Parameters
        ----------
        filepath : str
            Path where the hdf file has to be written.
        key : str or Group, optional
            Key (path) of the group within the HDF file (see Notes below).
            If None, the name of the group is used.
            Defaults to None.
        axis_key : str, optional
            Key (path) of the associated axis in the HDF file (see Notes below).
            If None, the name of the axis associated with the group is used.
            Defaults to None.

        Notes
        -----
        Objects stored in a HDF file can be grouped together in `HDF groups`.
        If an object 'my_obj' is stored in a HDF group 'my_group',
        the key associated with this object is then 'my_group/my_obj'.
        Be aware that a HDF group can have subgroups.

        Examples
        --------
        >>> from larray import Axis
        >>> a = Axis("a=a0..a2")
        >>> a.to_hdf('test.h5')                     # doctest: +SKIP
        >>> a01 = a['a0,a1'] >> 'a01'

        Save group

        >>> # by default, the key is the name of the group
        >>> # and axis_key the name of the associated axis
        >>> a01.to_hdf('test.h5')                   # doctest: +SKIP

        Save group with a specific key

        >>> a01.to_hdf('test.h5', 'a_01')           # doctest: +SKIP

        Save group in a specific HDF group

        >>> a.to_hdf('test.h5', 'groups/a01')       # doctest: +SKIP

        The associated axis is saved with the group if not already present in the HDF file

        >>> b = Axis("b=b0..b2")
        >>> b01 = b['b0,b1'] >> 'b01'
        >>> # save both the group 'b01' and the associated axis 'b'
        >>> b01.to_hdf('test.h5')                   # doctest: +SKIP
        """
        if key is None:
            if self.name is None:
                raise ValueError("Argument key must be provided explicitly in case of anonymous group")
            key = self.name
        key = _translate_group_key_hdf(key)
        if axis_key is None:
            if self.axis.name is None:
                raise ValueError("Argument axis_key must be provided explicitly if the associated axis is anonymous")
            axis_key = self.axis.name
        data = self.eval()
        dtype_kind = data.dtype.kind if isinstance(data, np.ndarray) else ''
        if dtype_kind == 'U':
            data = np.char.encode(data, 'utf-8')
        s = pd.Series(data=data, name=self.name)
        with LHDFStore(filepath) as store:
            store.put(key, s)
            store.get_storer(key).attrs.type = 'Group'
            store.get_storer(key).attrs.dtype_kind = dtype_kind
            if axis_key not in store:
                self.axis.to_hdf(store, key=axis_key)
            store.get_storer(key).attrs.axis_key = axis_key

    # this makes range(LGroup(int)) possible
    def __index__(self):
        return self.eval().__index__()

    def __int__(self):
        # 'str' objects have no '__int__' attribute, so this is better than calling __int__ explicitly
        return int(self.eval())

    def __float__(self):
        # 'str' objects have no '__float__' attribute, so this is better than calling __float__ explicitly
        return float(self.eval())

    def __array__(self, dtype=None):
        return np.asarray(self.eval(), dtype=dtype)

    def __dir__(self):
        # called by dir() and tab-completion at the interactive prompt, must return a list of any valid getattr key.
        # dir() takes care of sorting but not uniqueness, so we must ensure that.
        attributes = self.__slots__
        return list(set(dir(self.eval())) | set(attributes) | set(dir(self.__class__)))

    def __getattr__(self, key):
        if key == '__array_struct__':
            raise AttributeError("'Group' object has no attribute '__array_struct__'")
        else:
            return getattr(self.eval(), key)

    # needed to make *un*pickling work (because otherwise, __getattr__ is called before .key exists, which leads to
    # an infinite recursion)
    def __getstate__(self):
        return self.key, self.name, self.axis

    def __setstate__(self, d):
        self.key, self.name, self.axis = d

    def __hash__(self):
        # to_tick & to_key are partially opposite operations but this standardize on a single notation so that they can
        # all target each other. eg, this removes spaces in "list strings", instead of hashing them directly
        # XXX: but we might want to include that normalization feature in to_tick directly, instead of using to_key
        #      explicitly here
        # XXX: we might want to make hash use the position along the axis instead of the labels so that if an axis has
        #      ambiguous labels, they do not hash to the same thing.
        # XXX: for performance reasons, I think hash should not evaluate slices. It should only translate pos bounds to
        #      labels or vice versa. We would loose equality between list Groups and equivalent slice groups but that
        #      is a small price to pay if the performance impact is large.
        # the problem with using self.translate() is that we cannot compare groups without axis
        # return hash(_to_tick(self.translate()))
        return hash(_to_tick(self.key))


def remove_nested_groups(key):
    # "struct" key with Group elements -> key without Group
    # TODO: ideally if all key elements are groups on the same Axis, we should make a group on that axis
    #       for slice bounds, watch out for None
    if isinstance(key, slice):
        key_start, key_stop = key.start, key.stop
        start = key_start.to_label() if isinstance(key_start, Group) else key_start
        stop = key_stop.to_label() if isinstance(key_stop, Group) else key_stop
        return slice(start, stop, key.step)
    elif isinstance(key, (tuple, list)):
        res = [k.to_label() if isinstance(k, Group) else k for k in key]
        return tuple(res) if isinstance(key, tuple) else res
    else:
        return key


class LGroup(Group):
    r"""Label group.

    Represents a subset of labels of an axis.

    Parameters
    ----------
    key : key
        Anything usable for indexing. A key should be either sequence of labels, a slice with label bounds or a string.
    name : str, optional
        Name of the group.
    axis : int, str, Axis, optional
        Axis for group.

    Examples
    --------
    >>> from larray import Axis, X
    >>> age = Axis('0..100', 'age')
    >>> teens = X.age[10:19].named('teens')
    >>> teens
    X.age[10:19] >> 'teens'
    >>> teens = X.age[10:19] >> 'teens'
    >>> teens
    X.age[10:19] >> 'teens'
    """
    __slots__ = ()
    format_string = "{axis}[{key}]"

    def __init__(self, key, name=None, axis=None):
        key = _to_key(key)
        Group.__init__(self, key, name, axis)

    # XXX: return IGroup instead?
    def translate(self, bound=None, stop=False):
        r"""
        compute position(s) of group
        """
        if bound is None:
            bound = self.key
        if isinstance(self.axis, ABCAxis):
            pos = self.axis.index(bound)
            return (pos + int(stop)) if np.isscalar(pos) else pos
        else:
            raise ValueError("Cannot translate an LGroup without axis")

    def to_label(self):
        return self.key

    def eval(self):
        if isinstance(self.key, slice):
            if isinstance(self.axis, ABCAxis):
                # expand slices
                return self.axis.labels[self.translate()]
            else:
                return self.key
                # raise ValueError("Cannot evaluate a slice group without axis")
        else:
            # we do not check the group labels are actually valid on Axis
            return self.key


class LSet(LGroup):
    r"""Label set.

    Represents a set of (unique) labels of an axis.

    Parameters
    ----------
    key : key
        Anything usable for indexing. A key should be either sequence of labels, a slice with label bounds or a string.
    name : str, optional
        Name of the set.
    axis : int, str, Axis, optional
        Axis for set.

    Examples
    --------
    >>> from larray import Axis
    >>> letters = Axis('letters=a..z')
    >>> abc = letters[':c'].set() >> 'abc'
    >>> abc
    letters['a', 'b', 'c'].set() >> 'abc'
    >>> abc & letters['b:d']
    letters['b', 'c'].set()
    """
    __slots__ = ()
    format_string = "{axis}[{key}].set()"

    def __init__(self, key, name=None, axis=None):
        key = _to_key(key)
        if isinstance(key, Group):
            if name is None:
                name = key.name
            if axis is None:
                axis = key.axis
            if not isinstance(key, LSet):
                key = key.eval()
        if np.isscalar(key):
            key = [key]
        key = OrderedSet(key)
        LGroup.__init__(self, key, name, axis)

    # method factory
    def _binop(opname, c):
        op_fullname = f'__{opname}__'

        # TODO: implement this in a delayed fashion for reference axes
        def opmethod(self, other):
            if not isinstance(other, LSet):
                other = LSet(other)
            axis = self.axis if self.axis is not None else other.axis

            # setting a meaningful name is hard when either one has no name
            if self.name is not None and other.name is not None:
                name = f'{self.name} {c} {other.name}'
            else:
                name = None
            # TODO: implement this in a more efficient way for ndarray keys which can be large
            result_set = getattr(self.key, op_fullname)(other.key)
            return LSet(result_set, name=name, axis=axis)
        opmethod.__name__ = op_fullname
        return opmethod

    union = _binop('or', '|')
    __or__ = union

    intersection = _binop('and', '&')
    __and__ = intersection

    difference = _binop('sub', '-')
    __sub__ = difference


class IGroup(Group):
    r"""Index Group.

    Represents a subset of indices of an axis.

    Parameters
    ----------
    key : key
        Anything usable for indexing. A key should be either a single position, a sequence of positions, or a slice
        with integer bounds.
    name : str, optional
        Name of the group.
    axis : int, str, Axis, optional
        Axis for group.
    """
    __slots__ = ()
    format_string = "{axis}.i[{key}]"

    def translate(self, bound=None, stop=False):
        r"""
        compute position(s) of group
        """
        if bound is not None:
            return bound
        else:
            return self.key

    def to_label(self):
        if isinstance(self.axis, ABCAxis):
            labels = self.axis.labels
            key = self.key
            if isinstance(key, slice):
                start = labels[key.start] if key.start is not None else None
                # FIXME: this probably breaks for reverse slices
                # - 1 because IGroup slice stop is excluded while LGroup slice stop is included
                stop = labels[key.stop - 1] if key.stop is not None else None
                return slice(start, stop, key.step)
            else:
                # key is a single int or tuple/list/array of them
                return labels[key]
        else:
            raise ValueError("Cannot evaluate a positional group without axis")

    def eval(self):
        if isinstance(self.axis, ABCAxis):
            return self.axis.labels[self.key]
        else:
            raise ValueError("Cannot evaluate a positional group without axis")

    def __hash__(self):
        return hash(('IGroup', _to_tick(self.key)))


PGroup = renamed_to(IGroup, 'PGroup')
